"""Helpers for the asset queries"""

import warnings
from concurrent.futures import ThreadPoolExecutor
from itertools import repeat
from mimetypes import guess_extension
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import requests
from tenacity import retry
from tenacity.stop import stop_after_attempt
from tenacity.wait import wait_random

from kili.authentication import KiliAuth
from kili.exceptions import NotFound
from kili.graphql import QueryOptions
from kili.graphql.operations.project.queries import ProjectQuery, ProjectWhere

from .exceptions import MissingPropertyError


def get_download_assets_function(
    auth: KiliAuth,
    download_media: bool,
    fields: List[str],
    project_id: str,
    local_media_dir: Optional[str],
) -> Tuple[Optional[Callable], List[str]]:
    """Get the function to be called after each batch of asset query.

    The function is either None or MediaDownloader.download_assets().

    Also returns the fields to be queried, which may be modified
    if the jsonContent field is necessary.
    """
    if not download_media:
        return None, fields
    projects = list(
        ProjectQuery(auth.client, auth.ssl_verify)(
            ProjectWhere(project_id=project_id), ["inputType"], QueryOptions(disable_tqdm=True)
        )
    )
    if len(projects) == 0:
        raise NotFound(
            f"project ID: {project_id}. Maybe your KILI_API_KEY does not belong to a member of the"
            " project."
        )
    input_type = projects[0]["inputType"]
    jsoncontent_field_added = False
    if input_type in ("TEXT", "VIDEO") and "jsonContent" not in fields:
        fields = fields + ["jsonContent"]
        jsoncontent_field_added = True
    return (
        MediaDownloader(
            local_media_dir,
            project_id,
            jsoncontent_field_added,
            input_type,
            auth.ssl_verify,
        ).download_assets,
        fields,
    )


class MediaDownloader:
    """Media downloader for kili.assets()"""

    def __init__(  # pylint: disable=too-many-arguments
        self,
        local_media_dir: Optional[Union[Path, str]],
        project_id: str,
        jsoncontent_field_added: bool,
        project_input_type: str,
        ssl_verify: Union[bool, str],
    ) -> None:
        self.local_media_dir = local_media_dir
        self.project_id = project_id
        self.jsoncontent_field_added = jsoncontent_field_added
        self.project_input_type = project_input_type
        self.ssl_verify = ssl_verify

        self.local_dir_path = (
            Path(self.local_media_dir)
            if self.local_media_dir is not None
            else Path.home() / ".cache" / "kili" / "projects" / self.project_id / "assets"
        )

    def download_assets(self, assets: List[Dict]) -> List[Dict]:
        """Download assets media in local."""
        if len(assets) == 0:
            return assets

        assert_required_fields_existence(assets)

        self.local_dir_path.mkdir(parents=True, exist_ok=True)

        with ThreadPoolExecutor() as threads:
            assets_gen = threads.map(self.download_single_asset, assets)

        assets = list(assets_gen)

        if self.jsoncontent_field_added:
            jsoncontent_not_empty = any(bool(asset["jsonContent"]) for asset in assets)
            if jsoncontent_not_empty:
                warnings.warn(
                    "Non empty jsonContent found in assets. Field was automatically added.",
                    stacklevel=2,
                )
            else:
                for asset in assets:
                    del asset["jsonContent"]

        return assets

    def download_single_asset(self, asset: Dict) -> Dict[str, Any]:
        """Download single asset on disk and modify asset attributes"""

        if "jsonContent" in asset and str(asset["jsonContent"]).startswith("http"):
            # richtext
            if self.project_input_type == "TEXT":
                asset["jsonContent"] = download_file(
                    asset["jsonContent"], asset["externalId"], self.local_dir_path, self.ssl_verify
                )

            # video frames
            elif self.project_input_type == "VIDEO":
                urls = get_json_content_urls_video(asset["jsonContent"], self.ssl_verify)
                nbr_char_zfill = len(str(len(urls)))
                img_names = (
                    f'{asset["externalId"]}_{f"{i+1}".zfill(nbr_char_zfill)}'
                    for i, _ in enumerate(urls)
                )
                with ThreadPoolExecutor() as threads:
                    paths_gen = threads.map(
                        download_file,
                        urls,
                        img_names,
                        repeat(self.local_dir_path),
                        repeat(self.ssl_verify),
                    )
                    asset["jsonContent"] = list(paths_gen)
                return asset  # we skip video "content" download

            # big images
            elif self.project_input_type == "IMAGE":
                # the "jsonContent" contains some information but not the image
                response = requests.get(asset["jsonContent"], verify=self.ssl_verify, timeout=20)
                response = response.json()
                asset["jsonContent"] = response

            else:
                raise NotImplementedError(
                    f"jsonContent download for type {self.project_input_type} not implemented yet."
                )

        if str(asset["content"]).startswith("http"):
            asset["content"] = download_file(
                asset["content"],
                asset["externalId"],
                self.local_dir_path,
                ssl_verify=self.ssl_verify,
            )
            return asset

        return asset


def get_file_extension_from_headers(url, ssl_verify: Union[bool, str]) -> Optional[str]:
    """guess the extension of a file with the url response headers"""
    with requests.head(url, timeout=20, verify=ssl_verify) as header_response:
        if header_response.status_code == 200:
            headers = header_response.headers
        else:
            with requests.get(url, timeout=20, verify=ssl_verify) as response:
                response.raise_for_status()
                headers = response.headers
        if "content-type" in headers:
            content_type = headers["content-type"]
            return guess_extension(content_type)
    return None


def get_download_path(
    url: str, external_id: str, local_dir_path: Path, ssl_verify: Union[bool, str]
) -> Path:
    """Build the path to download a file the file in local."""
    extension = get_file_extension_from_headers(url, ssl_verify=ssl_verify)
    filename = external_id
    if extension is not None and not filename.endswith(extension):
        filename = filename + extension
    local_dir_path = local_dir_path / filename
    return local_dir_path.resolve()


@retry(stop=stop_after_attempt(2), wait=wait_random(min=1, max=2), reraise=True)
def download_file(
    url: str, external_id: str, local_dir_path: Path, ssl_verify: Union[bool, str]
) -> str:
    """
    Download a file by streming chunks of 1Mb
    If the file already exists in local, it does not download it
    """
    local_path = get_download_path(url, external_id, local_dir_path, ssl_verify)
    local_path.parent.mkdir(parents=True, exist_ok=True)
    if not local_path.is_file():
        with requests.get(url, stream=True, timeout=20, verify=ssl_verify) as response:
            response.raise_for_status()
            with open(local_path, "wb") as file:
                for chunk in response.iter_content(chunk_size=1024 * 1024):
                    file.write(chunk)
    return str(local_path)


def assert_required_fields_existence(assets: List[Dict]) -> None:
    """check if all fields are available to download assets"""
    required_fields = ["content", "externalId"]
    for field in required_fields:
        if field not in assets[0].keys():
            raise MissingPropertyError(
                f"The asset does not have the {field} field. Please add it to the fields when"
                " querying assets."
            )


def get_json_content_urls_video(json_url: str, ssl_verify: Union[bool, str]) -> Tuple[str]:
    """Get frame urls from a jsonContent url."""
    response = requests.get(json_url, timeout=20, verify=ssl_verify)
    response = response.json()
    urls = tuple(response.values())
    return urls



"""Service for exporting kili objects """

import warnings
from pathlib import Path
from typing import Dict, List, Optional, Type

from typing_extensions import get_args

from kili.authentication import KiliAuth
from kili.graphql.operations.asset.queries import AssetQuery, AssetWhere
from kili.services.export.format.base import AbstractExporter, ExportParams
from kili.services.export.format.coco import CocoExporter
from kili.services.export.format.kili import KiliExporter
from kili.services.export.format.voc import VocExporter
from kili.services.export.format.yolo import YoloExporter
from kili.services.export.logger import get_logger
from kili.services.export.repository import SDKContentRepository
from kili.services.export.types import (
    CocoAnnotationModifier,
    ExportType,
    LabelFormat,
    SplitOption,
)
from kili.services.project import get_project
from kili.services.types import LogLevel, ProjectId

THRESHOLD_WARN_MANY_ASSETS = 1000


def export_labels(  # pylint: disable=too-many-arguments, too-many-locals
    auth: KiliAuth,
    asset_ids: Optional[List[str]],
    project_id: ProjectId,
    export_type: ExportType,
    label_format: LabelFormat,
    split_option: SplitOption,
    single_file: bool,
    output_file: str,
    disable_tqdm: bool,
    log_level: LogLevel,
    with_assets: bool,
    annotation_modifier: Optional[CocoAnnotationModifier],
) -> None:
    """
    Export the selected assets into the required format, and save it into a file archive.
    """
    get_project(auth, project_id, ["id"])

    if with_assets:
        count = AssetQuery(auth.client, auth.ssl_verify).count(AssetWhere(project_id=project_id))
        if count > THRESHOLD_WARN_MANY_ASSETS:
            warnings.warn(
                (
                    f"Downloading many assets ({count}). This might take a while. Consider"
                    " disabling assets download in the options."
                ),
                stacklevel=2,
            )

    export_params = ExportParams(
        assets_ids=asset_ids,
        project_id=project_id,
        export_type=export_type,
        label_format=label_format,
        split_option=split_option,
        single_file=single_file,
        output_file=Path(output_file),
        with_assets=with_assets,
        annotation_modifier=annotation_modifier,
    )

    logger = get_logger(log_level)

    content_repository = SDKContentRepository(
        auth.api_endpoint,
        router_headers={
            "Authorization": f"X-API-Key: {auth.api_key}",
        },
        verify_ssl=True,
    )

    if label_format in get_args(LabelFormat):
        format_exporter_selector_mapping: Dict[str, Type[AbstractExporter]] = {
            "raw": KiliExporter,
            "kili": KiliExporter,
            "coco": CocoExporter,
            "yolo_v4": YoloExporter,
            "yolo_v5": YoloExporter,
            "yolo_v7": YoloExporter,
            "pascal_voc": VocExporter,
        }
        assert set(format_exporter_selector_mapping.keys()) == set(
            get_args(LabelFormat)
        )  # ensures full mapping
        exporter_class = format_exporter_selector_mapping[label_format]
        exporter_class(
            export_params, auth, logger, disable_tqdm, content_repository
        ).export_project()
    else:
        raise ValueError(f'Label format "{label_format}" is not implemented or does not exist.')



"""Service module for projects"""

from typing import List

from kili.authentication import KiliAuth
from kili.exceptions import NotFound
from kili.graphql import QueryOptions
from kili.graphql.operations.project.queries import ProjectQuery, ProjectWhere


def get_project(auth: KiliAuth, project_id: str, fields: List[str]):
    """Get a project from its id or raise a NotFound Error if not found"""
    projects = list(
        ProjectQuery(auth.client, auth.ssl_verify)(
            ProjectWhere(project_id=project_id), fields, QueryOptions(disable_tqdm=True)
        )
    )
    if len(projects) == 0:
        raise NotFound(
            f"project ID: {project_id}. Maybe your KILI_API_KEY does not belong to a member of the"
            " project."
        )
    return projects[0]


def get_project_field(auth: KiliAuth, project_id: str, field: str):
    """Get one project field from a the project id or raise a NotFound Error
    if the project is not found
    """
    return get_project(auth, project_id, [field])[field]



"""
Module for methods and classes that are for internal use by Kili Technology only.
"""

from typeguard import typechecked

from kili.helpers import format_result
from kili.mutations.project.queries import GQL_DELETE_PROJECT

from ..mutations.organization import MutationsOrganization
from ..mutations.user.queries import GQL_RESET_PASSWORD


class KiliInternal(MutationsOrganization):
    """Inherit classes for internal use by Kili Technology only."""

    def __init__(self, kili):
        """Initializes the class.

        Args:
            kili: Kili object
        """
        self.kili = kili
        super().__init__(self.kili.auth)

    @typechecked
    def reset_password(self, email: str):
        """Reset password.
        WARNING: This method is for internal use only.

        Args:
            email: Email of the person whose password has to be reset.

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.
        """
        variables = {"where": {"email": email}}
        result = self.auth.client.execute(GQL_RESET_PASSWORD, variables)
        return format_result("data", result, None, self.auth.ssl_verify)

    @typechecked
    def delete_project(self, project_id: str):
        """Delete project permanently.
        WARNING: This resolver is for internal use by Kili Technology only.

        Args:
            project_id: Identifier of the project

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.
        """
        variables = {"projectID": project_id}
        result = self.auth.client.execute(GQL_DELETE_PROJECT, variables)
        return format_result("data", result, None, self.auth.ssl_verify)



"""
Label Importers
"""
import csv
import logging
from abc import ABC, abstractmethod
from json import dumps
from pathlib import Path
from typing import Dict, List, NamedTuple, Optional, Type

import yaml

from kili.authentication import KiliAuth
from kili.graphql.operations.label.mutations import GQL_APPEND_MANY_LABELS
from kili.helpers import format_result, get_file_paths_to_upload
from kili.orm import Label
from kili.services.helpers import (
    get_external_id_from_file_path,
    infer_ids_from_external_ids,
)
from kili.services.label_import.exceptions import (
    LabelParsingError,
    MissingMetadataError,
    MissingTargetJobError,
)
from kili.services.label_import.parser import (
    AbstractLabelParser,
    KiliRawLabelParser,
    YoloLabelParser,
)
from kili.services.label_import.types import Classes, LabelFormat
from kili.services.types import LabelType, LogLevel, ProjectId
from kili.utils import pagination, tqdm


class LoggerParams(NamedTuple):
    """
    Contains all parameters related to logging.
    """

    disable_tqdm: bool
    level: LogLevel


class AbstractLabelImporter(ABC):
    """
    Abstract Label Importer
    """

    def __init__(self, auth: KiliAuth, logger_params: LoggerParams, input_format: LabelFormat):
        self.auth = auth
        self.logger_params = logger_params
        self.input_format: LabelFormat = input_format

        self.logger = logging.getLogger("kili.services.label_import")
        self.logger.setLevel(logger_params.level)

    def process_from_files(  # pylint: disable=too-many-arguments
        self,
        labels_files: List[Path],
        meta_file_path: Optional[Path],
        project_id: ProjectId,
        target_job_name: Optional[str],
        model_name: Optional[str],
        is_prediction: bool,
    ):
        """
        Performs the import from the label files
        """
        self._check_arguments_compatibility(meta_file_path, target_job_name)
        class_by_id = self._read_classes_from_meta_file(meta_file_path, self.input_format)
        label_parser_class = self._select_label_parser()
        label_parser = label_parser_class(class_by_id, target_job_name)

        self.logger.warning("Importing labels")
        labels = self.extract_from_files(labels_files, label_parser)
        label_type = "PREDICTION" if is_prediction else "DEFAULT"
        self.process_from_dict(
            project_id=project_id,
            labels=labels,
            label_type=label_type,
            model_name=model_name,
        )

        self.logger.warning(print(f"{len(labels)} labels have been successfully imported"))

    def process_from_dict(
        self,
        project_id: Optional[str],
        labels: List[Dict],
        label_type: LabelType,
        model_name: Optional[str] = None,
    ) -> List:
        """
        Imports labels from a list of dictionaries representing labels.
        """
        should_retrieve_asset_ids = labels[0].get("asset_id") is None
        if should_retrieve_asset_ids:
            assert project_id
            asset_external_ids = [label["asset_external_id"] for label in labels]
            asset_id_map = infer_ids_from_external_ids(self.auth, asset_external_ids, project_id)
            labels = [
                {**label, "asset_id": asset_id_map[label["asset_external_id"]]} for label in labels
            ]
        labels_data = [
            {
                "jsonResponse": dumps(label.get("json_response")),
                "assetID": label.get("asset_id"),
                "secondsToLabel": label.get("seconds_to_label"),
                "modelName": model_name,
                "authorID": label.get("author_id"),
            }
            for label in labels
        ]
        batch_generator = pagination.BatchIteratorBuilder(labels_data)
        result = []
        with tqdm.tqdm(total=len(labels_data), disable=self.logger_params.disable_tqdm) as pbar:
            for batch_labels in batch_generator:
                variables = {
                    "data": {"labelType": label_type, "labelsData": batch_labels},
                    "where": {"idIn": [label["assetID"] for label in batch_labels]},
                }
                # we increase the timeout because the import can take a long time
                batch_result = self.auth.client.execute(
                    GQL_APPEND_MANY_LABELS, variables, timeout=60
                )
                result.extend(format_result("data", batch_result, Label, self.auth.ssl_verify))
                pbar.update(len(batch_labels))
        return result

    @staticmethod
    @abstractmethod
    def _check_arguments_compatibility(
        meta_file_path: Optional[Path], target_job_name: Optional[str]
    ):
        pass

    @classmethod
    @abstractmethod
    def _read_classes_from_meta_file(
        cls, meta_file_path: Optional[Path], input_format: LabelFormat
    ) -> Optional[Classes]:
        pass

    @abstractmethod
    def _select_label_parser(self) -> Type[AbstractLabelParser]:
        pass

    @abstractmethod
    def _get_label_file_extension(self) -> str:
        pass

    def extract_from_files(
        self, labels_files: List[Path], label_parser: AbstractLabelParser
    ) -> List[Dict]:
        """
        Extracts the labels files and their metadata from the label files
        """

        label_paths = get_file_paths_to_upload(
            [str(f) for f in labels_files],
            lambda path: path.endswith(self._get_label_file_extension()),
            verbose=self.logger_params.level in ["INFO", "DEBUG"],
        )
        if len(label_paths) == 0:
            raise ValueError(
                "No label files to upload. Check that the paths exist and file types are .json"
            )
        labels = []
        for path in label_paths:
            try:
                labels.append(
                    {
                        "json_response": label_parser.parse(Path(path)),
                        "asset_external_id": get_external_id_from_file_path(Path(path)),
                    }
                )
            except Exception as exc:
                raise LabelParsingError(f"Failed to parse the file {path}") from exc
        return labels


class YoloLabelImporter(AbstractLabelImporter):
    """
    Label importer in the yolo format.
    """

    def _get_label_file_extension(self) -> str:
        return ".txt"

    @staticmethod
    def _check_arguments_compatibility(
        meta_file_path: Optional[Path], target_job_name: Optional[str]
    ):
        if meta_file_path is None:
            raise MissingMetadataError("Meta file is needed to import the label")
        if target_job_name is None:
            raise MissingTargetJobError("A target job name is needed to import the label")

    @staticmethod
    def _read_classes_from_meta_file(
        meta_file_path: Optional[Path], input_format: LabelFormat
    ) -> Classes:
        assert meta_file_path
        classes: Classes = Classes({})
        if input_format == "yolo_v4":
            try:
                # layout: id class\n
                with meta_file_path.open("r", encoding="utf-8") as m_f:
                    csv_reader = csv.reader(m_f, delimiter=" ")
                    classes = Classes({int(r[0]): r[1] for r in csv_reader if r[0] != " "})
            except (ValueError, IndexError):
                with meta_file_path.open("r", encoding="utf-8") as m_f:
                    classes = Classes(dict(enumerate(m_f.read().splitlines())))

        elif input_format == "yolo_v5":
            with meta_file_path.open("r", encoding="utf-8") as m_f:
                m_d = yaml.load(m_f, yaml.FullLoader)
                classes = Classes(m_d["names"])
        elif input_format == "yolo_v7":
            with meta_file_path.open("r", encoding="utf-8") as m_f:
                m_d = yaml.load(m_f, yaml.FullLoader)
                classes = Classes(dict(enumerate(m_d["names"])))
        else:
            raise NotImplementedError(f"The format f{input_format} does not have a metadata parser")

        return classes

    def _select_label_parser(self) -> Type[AbstractLabelParser]:
        return YoloLabelParser


class KiliRawLabelImporter(AbstractLabelImporter):
    """
    # Label importer in the Kili format.
    """

    def _get_label_file_extension(self) -> str:
        return ".json"

    @staticmethod
    def _check_arguments_compatibility(
        meta_file_path: Optional[Path], target_job_name: Optional[str]
    ):
        pass

    @classmethod
    def _read_classes_from_meta_file(
        cls, meta_file_path: Optional[Path], input_format: LabelFormat
    ) -> Optional[Classes]:
        return None

    def _select_label_parser(self) -> Type[AbstractLabelParser]:
        return KiliRawLabelParser



"""
Set of common functions used by different export formats
"""
from typing import Dict, List, Optional

from kili.authentication import KiliAuth
from kili.graphql import QueryOptions
from kili.graphql.operations.asset.queries import AssetQuery, AssetWhere
from kili.queries.asset.media_downloader import get_download_assets_function
from kili.services.export.exceptions import NotExportableAssetError
from kili.services.export.types import ExportType

DEFAULT_FIELDS = [
    "id",
    "content",
    "externalId",
    "jsonMetadata",
    "labels.author.id",
    "labels.author.email",
    "labels.author.firstname",
    "labels.author.lastname",
    "labels.jsonResponse",
    "labels.createdAt",
    "labels.isLatestLabelForUser",
    "labels.labelType",
    "labels.modelName",
]
LATEST_LABEL_FIELDS = [
    "id",
    "content",
    "externalId",
    "jsonContent",
    "jsonMetadata",
    "latestLabel.author.id",
    "latestLabel.author.email",
    "latestLabel.jsonResponse",
    "latestLabel.author.firstname",
    "latestLabel.author.lastname",
    "latestLabel.createdAt",
    "latestLabel.isLatestLabelForUser",
    "latestLabel.labelType",
    "latestLabel.modelName",
]


def attach_name_to_assets_labels_author(assets: List[Dict], export_type: ExportType):
    """
    Adds `name` field for author, by concatenating his/her first and last name
    """
    for asset in assets:
        if export_type == "latest":
            latest_label = asset["latestLabel"]
            if latest_label:
                firstname = latest_label["author"]["firstname"]
                lastname = latest_label["author"]["lastname"]
                latest_label["author"]["name"] = f"{firstname} {lastname}"
            continue
        for label in asset.get("labels", []):
            firstname = label["author"]["firstname"]
            lastname = label["author"]["lastname"]
            label["author"]["name"] = f"{firstname} {lastname}"


def fetch_assets(  # pylint: disable=too-many-arguments
    auth: KiliAuth,
    project_id: str,
    asset_ids: Optional[List[str]],
    export_type,
    label_type_in=None,
    disable_tqdm: bool = False,
    download_media: bool = False,
    local_media_dir: Optional[str] = None,
):
    """Fetches assets.

    Fetches assets where ID are in asset_ids if the list has more than one element,
    else all the assets of the project. If download media is passed, the media are
    downloaded into the `$HOME/.cache` folder.

    Args:
        project_id: project id
        assets_ids: list of asset IDs
        export_type: type of export (latest label or all labels)
        label_type_in: types of label to fetch (default, reviewed, ...)
        disable_tqdm: tell to disable tqdm
        download_media: tell to download the media in the cache folder.
        local_media_dir: Directory where the media are downloaded if `download_media` is True.

    Returns:
        List of fetched assets.
    """

    fields = get_fields_to_fetch(export_type)
    assets = None

    if asset_ids is not None and len(asset_ids) > 0:
        where = AssetWhere(
            asset_id_in=asset_ids,
            project_id=project_id,
            label_type_in=label_type_in,
        )
    else:
        where = AssetWhere(
            project_id=project_id,
            label_type_in=label_type_in,
        )
    options = QueryOptions(disable_tqdm=disable_tqdm)
    post_call_function, fields = get_download_assets_function(
        auth, download_media, fields, project_id, local_media_dir
    )
    assets = list(
        AssetQuery(auth.client, auth.ssl_verify)(where, fields, options, post_call_function)
    )
    _check_content_presence(assets)
    attach_name_to_assets_labels_author(assets, export_type)
    return assets


def _check_content_presence(assets: List[Dict]):
    if any(a.get("content") in [None, ""] and a.get("jsonContent") in [None, ""] for a in assets):
        raise NotExportableAssetError(
            "The assets cannot be exported. This can happen if they are hosted in a cloud storage."
        )


def get_fields_to_fetch(export_type: ExportType):
    """
    Returns the fields to fetch depending on the export type
    """
    if export_type == "latest":
        return LATEST_LABEL_FIELDS
    return DEFAULT_FIELDS



"""Data integration queries."""

from typing import Dict, Generator, Iterable, List, Optional, overload

from typeguard import typechecked
from typing_extensions import Literal

from kili.authentication import KiliAuth
from kili.graphql import QueryOptions
from kili.graphql.operations.data_integration.queries import (
    DataIntegrationsQuery,
    DataIntegrationWhere,
)
from kili.helpers import disable_tqdm_if_as_generator
from kili.utils.logcontext import for_all_methods, log_call


@for_all_methods(log_call, exclude=["__init__"])
class QueriesDataIntegration:
    """
    Set of cloud storage integration queries
    """

    # pylint: disable=too-many-arguments,dangerous-default-value

    def __init__(self, auth: KiliAuth):
        """Initialize the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    @overload
    def cloud_storage_integrations(
        self,
        cloud_storage_integration_id: Optional[str] = None,
        name: Optional[str] = None,
        platform: Optional[Literal["AWS", "Azure", "GCP"]] = None,
        status: Optional[Literal["CONNECTED", "DISCONNECTED", "CHECKING"]] = None,
        organization_id: Optional[str] = None,
        fields: List[str] = ["name", "id", "platform", "status"],
        first: Optional[int] = None,
        skip: int = 0,
        disable_tqdm: bool = False,
        *,
        as_generator: Literal[True],
    ) -> Generator[Dict, None, None]:
        ...

    @overload
    def cloud_storage_integrations(
        self,
        cloud_storage_integration_id: Optional[str] = None,
        name: Optional[str] = None,
        platform: Optional[Literal["AWS", "Azure", "GCP"]] = None,
        status: Optional[Literal["CONNECTED", "DISCONNECTED", "CHECKING"]] = None,
        organization_id: Optional[str] = None,
        fields: List[str] = ["name", "id", "platform", "status"],
        first: Optional[int] = None,
        skip: int = 0,
        disable_tqdm: bool = False,
        *,
        as_generator: Literal[False] = False,
    ) -> List[Dict]:
        ...

    @typechecked
    def cloud_storage_integrations(
        self,
        cloud_storage_integration_id: Optional[str] = None,
        name: Optional[str] = None,
        platform: Optional[Literal["AWS", "Azure", "GCP"]] = None,
        status: Optional[Literal["CONNECTED", "DISCONNECTED", "CHECKING"]] = None,
        organization_id: Optional[str] = None,
        fields: List[str] = ["name", "id", "platform", "status"],
        first: Optional[int] = None,
        skip: int = 0,
        disable_tqdm: bool = False,
        *,
        as_generator: bool = False,
    ) -> Iterable[Dict]:
        # pylint: disable=line-too-long
        """Get a generator or a list of cloud storage integrations that match a set of criteria.

        Args:
            cloud_storage_integration_id: ID of the cloud storage integration.
            name: Name of the cloud storage integration.
            platform: Platform of the cloud storage integration.
            status: Status of the cloud storage integration.
            organization_id: ID of the organization.
            fields: All the fields to request among the possible fields for the cloud storage integrations.
                See [the documentation](https://docs.kili-technology.com/reference/graphql-api#dataintegration) for all possible fields.
            first: Maximum number of cloud storage integrations to return.
            skip: Number of skipped cloud storage integrations.
            disable_tqdm: If `True`, the progress bar will be disabled.
            as_generator: If `True`, a generator on the cloud storage integrations is returned.

        Returns:
            A list or a generator of the cloud storage integrations that match the criteria.

        Examples:
            >>> kili.cloud_storage_integrations()
            [{'name': 'My bucket', 'id': '123456789', 'platform': 'AWS', 'status': 'CONNECTED'}]
        """
        where = DataIntegrationWhere(
            data_integration_id=cloud_storage_integration_id,
            name=name,
            platform=platform,
            status=status,
            organization_id=organization_id,
        )
        disable_tqdm = disable_tqdm_if_as_generator(as_generator, disable_tqdm)
        options = QueryOptions(disable_tqdm, first, skip)
        data_integrations_gen = DataIntegrationsQuery(self.auth.client, self.auth.ssl_verify)(
            where, fields, options
        )

        if as_generator:
            return data_integrations_gen
        return list(data_integrations_gen)

    @typechecked
    def count_cloud_storage_integrations(
        self,
        cloud_storage_integration_id: Optional[str] = None,
        name: Optional[str] = None,
        platform: Optional[Literal["AWS", "Azure", "GCP"]] = None,
        status: Optional[Literal["CONNECTED", "DISCONNECTED", "CHECKING"]] = None,
        organization_id: Optional[str] = None,
    ) -> int:
        """Count and return the number of cloud storage integrations that match a set of criteria.

        Args:
            cloud_storage_integration_id: ID of the cloud storage integration.
            name: Name of the cloud storage integration.
            platform: Platform of the cloud storage integration.
            status: Status of the cloud storage integration.
            organization_id: ID of the organization.

        Returns:
            The number of cloud storage integrations that match the criteria.
        """
        where = DataIntegrationWhere(
            data_integration_id=cloud_storage_integration_id,
            name=name,
            platform=platform,
            status=status,
            organization_id=organization_id,
        )
        return DataIntegrationsQuery(self.auth.client, self.auth.ssl_verify).count(where)



"""Tools for export with images."""

from pathlib import Path
from typing import Tuple, Union

from PIL import Image


def get_image_dimensions(file_path: Union[Path, str]) -> Tuple:
    """
    Get an image width and height
    """
    assert Path(file_path).is_file(), f"File {file_path} does not exist"
    image = Image.open(str(file_path))
    return image.size



"""
GraphQL Queries of Locks
"""

from typing import Optional

from kili.graphql import BaseQueryWhere, GraphQLQuery


class LockWhere(BaseQueryWhere):
    """
    Tuple to be passed to the LockQuery to restrict query
    """

    def __init__(
        self,
        lock_id: Optional[str] = None,
    ):
        self.lock_id = lock_id
        super().__init__()

    def graphql_where_builder(self):
        """Build the GraphQL Where payload sent in the resolver from the SDK LockWhere"""
        return {"id": self.lock_id}


class LockQuery(GraphQLQuery):
    """Lock query."""

    @staticmethod
    def query(fragment):
        """
        Return the GraphQL locks query
        """
        return f"""
        query locks($where: LockWhere!, $first: PageSize!, $skip: Int!) {{
            data: locks(where: $where, first: $first, skip: $skip) {{
                {fragment}
            }}
        }}
        """

    COUNT_QUERY = """
    query countLocks($where: LockWhere!) {
        data: countLocks(where: $where)
    }
    """



"""
GraphQL Queries of API keys
"""


from typing import Optional

from kili.graphql import BaseQueryWhere, GraphQLQuery


class APIKeyWhere(BaseQueryWhere):
    """
    Tuple to be passed to the APIKeyQuery to restrict the query
    """

    def __init__(
        self,
        api_key_id: Optional[str] = None,
        user_id: Optional[str] = None,
        api_key: Optional[str] = None,
    ):
        self.api_key_id = api_key_id
        self.user_id = user_id
        self.api_key = api_key
        super().__init__()

    def graphql_where_builder(self):
        """Build the GraphQL Where payload sent in the resolver from the SDK APIKeyWhere"""
        return {
            "user": {"id": self.user_id},
            "id": self.api_key_id,
            "key": self.api_key,
        }


class APIKeyQuery(GraphQLQuery):
    """APIKey query."""

    @staticmethod
    def query(fragment):
        """
        Return the GraphQL apiKeys query
        """
        return f"""
        query apiKeys($where: ApiKeyWhere!, $first: PageSize!, $skip: Int!) {{
          data: apiKeys(where: $where, skip: $skip, first: $first) {{
            {fragment}
          }}
        }}
        """

    COUNT_QUERY = """
    query countApiKeys($where: ApiKeyWhere!) {
      data: countApiKeys(where: $where)
    }
    """



"""Plugins queries."""

import json
from datetime import datetime
from typing import List, Optional

from typeguard import typechecked

from kili.graphql import QueryOptions
from kili.graphql.operations.plugin.queries import PluginLogsWhere, PluginQuery
from kili.services.plugins import PluginUploader
from kili.utils.logcontext import for_all_methods, log_call

from ...authentication import KiliAuth


@for_all_methods(log_call, exclude=["__init__"])
class QueriesPlugins:
    """Set of Plugins queries."""

    # pylint: disable=too-many-arguments

    def __init__(self, auth: KiliAuth):
        """Initialize the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    @typechecked
    def get_plugin_logs(
        self,
        project_id: str,
        plugin_name: str,
        start_date: Optional[datetime] = None,
        limit: Optional[int] = None,
        skip: Optional[int] = None,
    ):
        # pylint: disable=line-too-long
        """Get paginated logs of a plugin on a project.
        Args:
            project_id: Identifier of the project
            plugin_name: Name of the plugin
            start_date: Datetime used to get the logs from, if not provided, it will be the plugin's creation date
            limit: Limit for pagination, if not provided, it will be 100
            skip: Skip for pagination, if not provided, it will be 0
        Returns:
            A result array which contains the logs of the plugin,
                or an error message.
        Examples:
            >>> kili.get_plugin_logs(project_id="my_project_id", plugin_name="my_plugin_name", start_date="1970/01/01")
        """

        where = PluginLogsWhere(
            project_id=project_id, plugin_name=plugin_name, start_date=start_date
        )
        options = QueryOptions(
            first=limit, skip=skip or 0, disable_tqdm=False
        )  # disable tqm is not implemnted for this query
        pretty_result = PluginQuery(self.auth.client, self.auth.ssl_verify).get_logs(where, options)
        return json.dumps(pretty_result, sort_keys=True, indent=4)

    @typechecked
    def get_plugin_status(
        self,
        plugin_name: str,
        verbose: bool = True,
    ):
        """Update a plugin with new code.

        Args:
            plugin_name: Name of the plugin
            verbose: If false, minimal logs are displayed

        Returns:
            The status of the plugin if query was successful or an error message otherwise.

        Examples:
            >>> kili.get_plugin_status(plugin_name="my_plugin_name")
        """

        result = PluginUploader(self.auth, "", plugin_name, verbose).get_plugin_runner_status()
        return result

    # pylint: disable=dangerous-default-value
    @typechecked
    def list_plugins(
        self,
        fields: List[str] = [
            "name",
            "projectIds",
            "id",
            "createdAt",
            "updatedAt",
        ],
    ):
        # pylint: disable=line-too-long
        """List all plugins from your organization
        Args:
            fields: All the fields to request among the possible fields for the plugins
                See [the documentation](https://docs.kili-technology.com/reference/graphql-api#plugins) for all possible fields.

        Returns:
            A result array which contains all the plugins from your organization,
                or an error message.
        Examples:
            >>> kili.list_plugins()
            >>> kili.list_plugins(fields=['name'])
        """
        return PluginQuery(self.auth.client, self.auth.ssl_verify).list(fields=fields)



"""
GraphQL Queries of Notifications
"""


from typing import Optional

from kili.graphql import BaseQueryWhere, GraphQLQuery


class NotificationWhere(BaseQueryWhere):
    """
    Tuple to be passed to the NotificationQuery to restrict query
    """

    def __init__(
        self,
        has_been_seen: Optional[bool] = None,
        notification_id: Optional[str] = None,
        user_id: Optional[str] = None,
    ):
        self.has_been_seen = has_been_seen
        self.notification_id = notification_id
        self.user_id = user_id
        super().__init__()

    def graphql_where_builder(self):
        """Build the GraphQL Where payload sent in the resolver from the SDK NotificationWhere"""
        return {
            "id": self.notification_id,
            "user": {
                "id": self.user_id,
            },
            "hasBeenSeen": self.has_been_seen,
        }


class NotificationQuery(GraphQLQuery):
    """Notification query."""

    @staticmethod
    def query(fragment):
        """
        Return the GraphQL notifications query
        """

        return f"""
        query notifications($where: NotificationWhere!, $first: PageSize!, $skip: Int!) {{
            data: notifications(where: $where, first: $first, skip: $skip) {{
                {fragment}
            }}
        }}
        """

    COUNT_QUERY = """
    query countNotifications($where: NotificationWhere!) {
        data: countNotifications(where: $where)
    }
    """



"""
Functions to import assets into a PDF project
"""
from typing import List

from .base import BaseAssetImporter, BatchParams, ContentBatchImporter
from .types import AssetLike


class PdfDataImporter(BaseAssetImporter):
    """
    class for importing data into a PDF project
    """

    def import_assets(self, assets: List[AssetLike]):
        """
        Import PDF assets into Kili.
        """
        self._check_upload_is_allowed(assets)
        is_hosted = self.is_hosted_content(assets)
        assets = self.filter_duplicate_external_ids(assets)

        batch_params = BatchParams(is_hosted=is_hosted, is_asynchronous=False)
        batch_importer = ContentBatchImporter(
            self.auth, self.project_params, batch_params, self.pbar
        )
        result = self.import_assets_by_batch(assets, batch_importer)
        return result



"""CLI's project create subcommand"""

import json
from typing import Optional

import click
from tabulate import tabulate

from kili import services
from kili.cli.common_args import Options
from kili.cli.helpers import get_kili_client
from kili.constants import INPUT_TYPE
from kili.queries.project.helpers import get_project_url


@click.command(name="create")
@Options.api_key
@Options.endpoint
@click.argument("interface", type=click.Path(exists=True), required=False)
@Options.from_project
@click.option("--title", type=str, required=True, help="Project Title.")
@click.option(
    "--input-type",
    type=click.Choice(INPUT_TYPE),
    required=True,
    help="Project input data type. Please check your license to see which ones you have access to.",
)
@click.option("--description", type=str, default="", help="Project description.")
@Options.tablefmt
# pylint: disable=too-many-arguments
def create_project(
    api_key: Optional[str],
    endpoint: Optional[str],
    interface: str,
    project_id_src: str,
    input_type,
    title: str,
    description: str,
    tablefmt: str,
    ssl_verify: str,
):
    """Create a Kili project

    interface must be a path pointing to your json interface file

    If no interface is provided, --from-project can be used
    to create a new project with the json_interface of another project
    (assets will not be copied).

    \b
    !!! Examples
        ```
        kili project create \\
             path/to/interface.json \\
            --input-type TEXT \\
            --title "Invoice annotation project"
        ```
        ```
        kili project create \\
            --from-project <project_id_src> \\
            --input-type TEXT \\
            --title "Invoice annotation project"
        ```
    To build a Kili project interface, please visit: \n
    https://docs.kili-technology.com/docs/customizing-the-interface-through-json-settings
    """
    kili = get_kili_client(api_key=api_key, api_endpoint=endpoint, ssl_verify=ssl_verify)

    if ((interface is not None) + (project_id_src is not None)) > 1:
        raise ValueError("interface argument and option --from-project are exclusive.")
    if ((interface is not None) + (project_id_src is not None)) == 0:
        raise ValueError("You must use either interface argument or option --from-project")

    if interface is not None:
        with open(interface, encoding="utf-8", mode="r") as interface_file:
            json_interface = json.load(interface_file)

    elif project_id_src is not None:
        json_interface = services.get_project_field(kili.auth, project_id_src, "jsonInterface")

    result = kili.create_project(
        input_type=input_type,
        json_interface=json_interface,
        title=title,
        description=description,
    )
    project_id = result["id"]

    project_url = get_project_url(project_id, kili.auth.client.endpoint)
    print(tabulate([[project_id, project_url]], headers=["ID", "URL"], tablefmt=tablefmt))



"""
Queries of issue mutations
"""

from .fragments import ISSUE_FRAGMENT

GQL_APPEND_TO_ISSUES = f"""
mutation(
    $data: AppendToIssuesData!
    $where: AssetWhere!
) {{
  data: appendToIssues(
    data: $data
    where: $where
  ) {{
    {ISSUE_FRAGMENT}
  }}
}}
"""

GQL_CREATE_ISSUES = f"""
mutation createIssues(
    $issues: [IssueToCreate!]!
    $where: AssetWhere!
) {{
  data: createIssues(
    issues: $issues
    where: $where
  ) {{
    {ISSUE_FRAGMENT}
  }}
}}
"""



"""Project module."""
from typing import List, Optional, cast

from typeguard import typechecked

from kili import services
from kili.services.export.types import LabelFormat, SplitOption
from kili.services.types import AssetId, InputType, LabelType, LogLevel, ProjectId

from ..helpers import deprecate


class Project:
    """
    Object that represents a project in Kili.
    It allows management operations such as uploading assets, uploading predictions,
    modifying the project's queue etc.
    It also allows queries from this project such as its assets, labels etc.
    """

    @deprecate(
        msg="This class is deprecated. Use `kili.projects(project_id='<MY_PROJECT_ID>')` instead",
        removed_in="2.131",
    )
    def __init__(self, project_id: ProjectId, input_type: InputType, title: str, client):
        """Initialize the class."""
        self.project_id = project_id
        self.title = title
        self.input_type = input_type
        self.client = client

    @deprecate(
        msg="This method is deprecated. Use `kili.export_labels()` instead",
        removed_in="2.131",
    )
    @typechecked
    def export(  # pylint: disable=too-many-arguments
        self,
        filename: str,
        fmt: LabelFormat,
        asset_ids: Optional[List[AssetId]] = None,
        layout: SplitOption = "split",
        single_file: bool = False,
        disable_tqdm: bool = False,
        log_level: LogLevel = "INFO",
        with_assets: bool = True,
    ) -> None:
        """Export the project assets with the requested format into the requested output path.

        Usage:
        ```
        from kili.client import Kili
        kili = Kili()
        project = kili.get_project("your_project_id")
        project.export("export.zip", fmt="yolo_v4")
        ```

        Args:
            filename: Relative or full path of the archive that will contain
                the exported data.
            fmt: Format of the exported labels.
            asset_ids: Optional list of the assets from which to export the labels.
            layout: Layout of the exported files: "split" means there is one folder
                per job, "merged" that there is one folder with every labels.
            single_file: Layout of the exported labels. Single file mode is
                only available for some specific formats (COCO and Kili).
            disable_tqdm: Disable the progress bar if True.
            log_level: Level of debugging.
            with_assets: Download the assets in the export.
        """
        services.export_labels(
            self.client,
            asset_ids=cast(Optional[List[str]], asset_ids),
            project_id=self.project_id,
            export_type="latest",
            label_format=fmt,
            split_option=layout,
            single_file=single_file,
            output_file=filename,
            disable_tqdm=disable_tqdm,
            log_level=log_level,
            with_assets=with_assets,
            annotation_modifier=None,
        )

    @deprecate(
        msg="This method is deprecated. Use `kili.append_labels()` instead",
        removed_in="2.131",
    )
    @typechecked
    def append_labels(
        self,
        labels: List[dict],
        label_type: LabelType = "DEFAULT",
        model_name: Optional[str] = None,
    ):
        """Append labels to assets.

        !!! info "fields of labels to append"
            Either provide an asset_id or an external_id
            ```
            class LabelData:
                asset_id: str
                asset_external_id: str
                json_response: Required[Dict]
                author_id: str
                seconds_to_label: int
            ```

        Args:
            labels: list of dictionnaries with informations about the labels to create.
            label_type: Can be one of `AUTOSAVE`, `DEFAULT`, `PREDICTION`, `REVIEW` or `INFERENCE`

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.

        Examples:
            >>> kili.append_to_labels(
                    [
                        {'json_response': {...}, asset_id: 'cl9wmlkuc00050qsz6ut39g8h'},
                        {'json_response': {...}, asset_id: 'cl9wmlkuw00080qsz2kqh8aiy'}
                    ]
                )

        """
        return services.import_labels_from_dict(
            self.client, self.project_id, labels, label_type, model_name
        )



"""
Functions to activate/deactivate a plugin
"""


from kili.authentication import KiliAuth
from kili.graphql.operations.plugin.mutations import (
    GQL_ACTIVATE_PLUGIN_ON_PROJECT,
    GQL_DEACTIVATE_PLUGIN_ON_PROJECT,
)
from kili.helpers import format_result
from kili.services.plugins.helpers import get_logger
from kili.services.plugins.tools import check_errors_plugin_activation


def activate_plugin(auth: KiliAuth, plugin_name: str, project_id: str):
    """
    Create a plugin in Kili
    """

    logger = get_logger()

    variables = {"pluginName": plugin_name, "projectId": project_id}

    result = auth.client.execute(GQL_ACTIVATE_PLUGIN_ON_PROJECT, variables)

    has_failed, already_activated = check_errors_plugin_activation(result, plugin_name, project_id)

    if not already_activated and not has_failed:
        logger.info(f'Plugin with name "{plugin_name}" activated on project "{project_id}"')

    return format_result("data", result, None, auth.ssl_verify) if not already_activated else None


def deactivate_plugin(auth: KiliAuth, plugin_name: str, project_id: str):
    """
    Create a plugin in Kili
    """

    logger = get_logger()

    variables = {"pluginName": plugin_name, "projectId": project_id}

    result = auth.client.execute(GQL_DEACTIVATE_PLUGIN_ON_PROJECT, variables)

    logger.info(f"Plugin {plugin_name} deactivated on project {project_id}")

    return format_result("data", result, None, auth.ssl_verify)



"""
GraphQL Queries of data connections
"""


from typing import Optional

from kili.graphql import BaseQueryWhere, GraphQLQuery


class DataConnectionsWhere(BaseQueryWhere):
    """
    Tuple to be passed to the DataConnectionsQuery to restrict the query
    """

    def __init__(
        self,
        project_id: Optional[str] = None,
        data_integration_id: Optional[str] = None,
    ) -> None:
        self.project_id = project_id
        self.data_integration_id = data_integration_id
        super().__init__()

    def graphql_where_builder(self):
        """Build the GraphQL Where payload sent in the resolver from the SDK DataConnectionsWhere"""
        return {
            "projectId": self.project_id,
            "integrationId": self.data_integration_id,
        }


class DataConnectionsQuery(GraphQLQuery):
    """DataConnections query."""

    @staticmethod
    def query(fragment):
        """
        Return the GraphQL dataConnections query
        """
        return f"""
        query dataConnections($where: DataConnectionsWhere!, $first: PageSize!, $skip: Int!) {{
          data: dataConnections(where: $where, first: $first, skip: $skip) {{
            {fragment}
          }}
        }}
        """


class DataConnectionIdWhere(BaseQueryWhere):
    """
    Tuple to be passed to the DataConnectionQuery to restrict the query
    """

    def __init__(
        self,
        data_connection_id: str,
    ) -> None:
        self.data_connection_id = data_connection_id
        super().__init__()

    def graphql_where_builder(self):
        """Build the GraphQL Where payload sent in the resolver from the SDK DataConnectionWhere"""
        return {
            "id": self.data_connection_id,
        }


class DataConnectionQuery(GraphQLQuery):
    """DataConnection query."""

    @staticmethod
    def query(fragment):
        """
        Return the GraphQL dataConnection query
        """
        return f"""
        query dataConnection($where: DataConnectionIdWhere!) {{
          data: dataConnection(where: $where) {{
            {fragment}
          }}
        }}
        """



"""
Data integration queries.
"""

GQL_GET_DATA_INTEGRATION_FOLDER_AND_SUBFOLDERS = """
query getFoldersAndSubfolders($dataIntegrationId: ID!, $rootFolder: String) {
  data: getFoldersAndSubfolders(dataIntegrationId: $dataIntegrationId, rootFolder: $rootFolder)
}
"""



"""API authentication module"""
import warnings
from datetime import datetime, timedelta
from typing import Dict, Union

import requests

from kili import __version__
from kili.graphql import QueryOptions
from kili.graphql.graphql_client import GraphQLClient, GraphQLClientName
from kili.graphql.operations.api_key.queries import APIKeyQuery, APIKeyWhere
from kili.graphql.operations.user.queries import GQL_ME
from kili.helpers import format_result

from .exceptions import AuthenticationFailed, UserNotFoundError

warnings.filterwarnings("default", module="kili", category=DeprecationWarning)


def get_version_without_patch(version):
    """Return the version of Kili API removing the patch version.

    Args:
        version
    """
    return ".".join(version.split(".")[:-1])


# pylint: disable=too-many-instance-attributes
class KiliAuth:
    """
    Kili authentication class
    """

    def __init__(
        self,
        api_key: str,
        api_endpoint: str,
        client_name: GraphQLClientName,
        ssl_verify: Union[bool, str],
    ) -> None:
        self.api_key = api_key
        self.api_endpoint = api_endpoint
        self.client_name = client_name
        self.ssl_verify = ssl_verify

        try:
            self.endpoint_kili_version = self.check_versions_match()
        except:  # pylint: disable=bare-except
            message = (
                "We could not check the version, there might be a version"
                "mismatch or the app might be in deployment"
            )
            warnings.warn(message, UserWarning, stacklevel=2)

        self.check_api_key_valid()

        self.client = GraphQLClient(
            endpoint=api_endpoint,
            api_key=api_key,
            client_name=client_name,
            ssl_verify=self.ssl_verify,
        )

        self.check_expiry_of_key_is_close()

        user = self.get_user()
        self.user_id = user["id"]
        self.user_email = user["email"]

    def check_versions_match(self) -> str:
        """Check that the versions of Kili Python SDK and Kili API are the same

        Args:
            api_endpoint: url of the Kili API
        """
        url = self.api_endpoint.replace("/graphql", "/version")
        response = requests.get(url, verify=self.ssl_verify, timeout=30).json()
        version = response["version"]
        if get_version_without_patch(version) != get_version_without_patch(__version__):
            message = (
                "Kili Python SDK version should match with Kili API version.\n"
                + f'Please install version: "pip install kili=={version}"'
            )
            warnings.warn(message, UserWarning, stacklevel=2)
        return version

    def check_api_key_valid(self) -> None:
        """Check that the api_key provided is valid"""
        response = requests.post(
            url=self.api_endpoint,
            data='{"query":"{ me { id email } }"}',
            verify=self.ssl_verify,
            timeout=30,
            headers={
                "Authorization": f"X-API-Key: {self.api_key}",
                "Accept": "application/json",
                "Content-Type": "application/json",
                "apollographql-client-name": self.client_name.value,
                "apollographql-client-version": __version__,
            },
        )
        if response.status_code == 200 and "email" in response.text and "id" in response.text:
            return

        raise AuthenticationFailed(
            api_key=self.api_key,
            api_endpoint=self.api_endpoint,
            error_msg=(
                "Cannot check API key validity: status_code"
                f" {response.status_code}\n\n{response.text}"
            ),
        )

    def check_expiry_of_key_is_close(self) -> None:
        """Check that the expiration date of the api_key is not too close.

        Args:
            api_key: key used to connect to the Kili API
        """
        warn_days = 30

        api_keys = APIKeyQuery(self.client, self.ssl_verify)(
            fields=["expiryDate"],
            where=APIKeyWhere(api_key=self.api_key),
            options=QueryOptions(disable_tqdm=True),
        )

        key_expiry = datetime.strptime(next(api_keys)["expiryDate"], r"%Y-%m-%dT%H:%M:%S.%fZ")
        key_remaining_time = key_expiry - datetime.now()
        key_soon_deprecated = key_remaining_time < timedelta(days=warn_days)
        if key_soon_deprecated:
            message = f"""
                Your api key will be deprecated on {key_expiry:%Y-%m-%d}.
                You should generate a new one on My account > API KEY."""
            warnings.warn(message, UserWarning, stacklevel=2)

    def get_user(self) -> Dict:
        """Get the current user from the api_key provided"""
        result = self.client.execute(GQL_ME)
        user = format_result("data", result, None, self.ssl_verify)
        if user is None or user["id"] is None or user["email"] is None:
            raise UserNotFoundError("No user attached to the API key was found")
        return user



# pylint: disable:invalid-name,too-many-instance-attributes

"""
GraphQL types
"""

from typing import Dict, List

from typing_extensions import TypedDict

from kili.enums import (
    InputType,
    InvitationStatus,
    IssueStatus,
    LabelType,
    LicenseType,
    LockType,
    NotificationStatus,
    OrganizationRole,
    ProjectRole,
    Right,
    Status,
)
from kili.helpers import deprecate


@deprecate()
class License(TypedDict):
    """
    A Wrapper for License GraphQL object.
    """

    api: bool
    apiPriority: bool
    canUsePlugins: bool
    enableSmartTools: bool
    expiryDate: str
    inputType: InputType
    maxNumberOfAnnotations: int
    maxNumberOfAnnotationsPerMonth: int
    maxNumberOfLabeledAssets: int
    maxNumberOfPredictionsPerMonth: int
    organizationId: int
    seats: int
    startDate: str
    type: LicenseType
    uploadCloudData: bool
    uploadLocalData: bool


@deprecate()
class OrganizationWithoutUser(TypedDict):
    """
    A wrapper for Organization GraphQL object.
    Defined in two steps to avoid cyclical dependencies.
    """

    id: str
    address: str
    canSeeDataset: bool
    city: str
    country: str
    license: License
    name: str
    numberOfAnnotations: int
    numberOfLabeledAssets: int
    numberOfHours: float
    zipCode: str


@deprecate()
class UserWithoutProjectUsers(TypedDict):
    """
    A wrapper for User GraphQL object.
    """

    id: str
    activated: bool
    createdAt: str
    email: str
    firstname: str
    lastname: str
    organization: OrganizationWithoutUser
    organizationId: str
    organizationRole: OrganizationRole
    rights: Right
    updatedAt: str


@deprecate()
class Organization(OrganizationWithoutUser):
    """
    A wrapper for Organization GraphQL object.
    """

    users: UserWithoutProjectUsers


@deprecate()
class ProjectUserWithoutProject(TypedDict):
    """
    A wrapper for ProjectUser GraphQL object.
    Defined in two steps to avoid cyclical dependencies.
    """

    id: str
    activated: bool
    consensusMark: float
    consensusMarkPerCategory: Dict
    consensusMarkCompute: float
    honeypotMark: float
    honeypotMarkCompute: float
    invitationStatus: InvitationStatus
    lastLabelingAt: str
    lastLabelingAtCompute: str
    numberOfAnnotations: int
    numberOfAnnotationsCompute: int
    numberOfLabeledAssets: int
    numberOfLabels: int
    numberOfLabelsCompute: int
    numberOfLabeledAssets: int
    numberOfLabeledAssetsCompute: int
    role: ProjectRole
    starred: bool
    totalDuration: float
    totalDurationCompute: float
    user: UserWithoutProjectUsers


@deprecate()
class DataConnection(TypedDict):
    """
    A wrapper for DataConnection GraphQL object.
    """

    id: str
    isChecking: bool
    isApplyingDataDifferences: bool
    numberOfAssets: int
    projectId: str
    dataIntegrationId: str


@deprecate()
class ProjectWithoutDataset(TypedDict, total=False):
    """
    A wrapper for Project GraphQL object.
    Defined in two steps to avoid cyclical dependencies.
    """

    id: str
    assetMetadata: Dict
    assetMetadataCompute: Dict
    author: UserWithoutProjectUsers
    canNavigateBetweenAssets: bool
    canSkipAsset: bool
    consensusMark: float
    consensusMarkPerCategory: Dict
    consensusTotCoverage: int
    createdAt: str
    dataConnections: DataConnection
    description: str
    honeypotMark: float
    inputType: InputType
    instructions: str
    interface: Dict
    interfaceCompute: Dict
    jsonInterface: Dict
    metadataTypes: Dict
    minConsensusSize: int
    mlTasks: str
    mlTasksCompute: str
    numberOfRemainingAssets: int
    numberOfAssets: int
    numberOfSkippedAssets: int
    numberOfOpenIssues: int
    numberOfOpenQuestions: int
    numberOfSolvedIssues: int
    numberOfSolvedQuestions: int
    numberOfReviewedAssets: int
    readPermissionsForAssetsAndLabels: str
    reviewCoverage: int
    rights: Right
    roles: ProjectUserWithoutProject
    shouldRelaunchKpiComputation: bool
    title: str
    updatedAt: str
    useHoneyPot: bool


@deprecate()
class ProjectUser(ProjectUserWithoutProject):
    """
    A wrapper for ProjectUser GraphQL object.
    """

    project: ProjectWithoutDataset


@deprecate()
class UserWithoutApiKey(UserWithoutProjectUsers):
    """
    A wrapper for User GraphQL object.
    """

    projectUsers: ProjectUser


@deprecate()
class ApiKey(TypedDict):
    """
    A wrapper for ApiKey GraphQL object.
    """

    createdAt: str
    id: str
    key: str
    name: str
    revoked: bool
    user: UserWithoutApiKey  # ? Why not just User
    userId: str


@deprecate()
class User(UserWithoutApiKey):
    """
    A wrapper for User GraphQL object.
    """

    apiKeys: ApiKey


@deprecate()
class LabelWithoutLabelOf(TypedDict):
    """
    A wrapper for Label GraphQL object.
    Defined in two steps to avoid cyclical dependencies.
    """

    id: str
    assetIdCompute: str
    authorIdCompute: str
    author: User
    createdAt: str
    honeypotMark: float
    honeypotMarkCompute: float
    inferenceMark: float
    inferenceMarkCompute: float
    isLatestLabelForUser: bool
    isLatestLabelForUserCompute: bool
    isLatestDefaultLabelForUser: bool
    isLatestDefaultLabelForUserCompute: bool
    isLatestReviewLabelForUser: bool
    isLatestReviewLabelForUserCompute: bool
    jsonResponse: Dict
    labelType: LabelType
    modelName: str
    numberOfAnnotations: int
    numberOfAnnotationsCompute: int
    projectIdCompute: str
    responseCompute: Dict
    searchCompute: Dict
    secondsToLabel: int
    totalSecondsToLabel: float
    totalSecondsToLabelCompute: float


@deprecate()
class Lock(TypedDict):
    """
    A wrapper for Lock GraphQL object.
    """

    id: str
    author: User
    authorIdCompute: str
    createdAt: str
    lockType: LockType
    lockOfIdCompute: str


@deprecate()
class CommentsWithoutCommentsOf(TypedDict):
    """
    A wrapper for Comment GraphQL object.
    Defined in two steps to avoid cyclical dependencies.
    """

    id: str
    author: ProjectUser
    createdAt: str
    text: str
    updatedAt: str


@deprecate()
class CommentWithoutIssue(TypedDict):
    """
    A wrapper for Comment GraphQL object.
    """

    id: str
    author: ProjectUser
    authorId: str
    createdAt: str
    issueId: str
    text: str
    updatedAt: str


@deprecate()
class IssueWithoutAsset(TypedDict):
    """
    A wrapper for Issue GraphQL object.
    Defined in two steps to avoid cyclical dependencies.
    """

    id: str
    assetId: str
    assignee: ProjectUser
    assigneeId: str
    author: ProjectUser
    authorId: str
    comments: CommentWithoutIssue
    createdAt: str
    hasBeenSeen: bool
    issueNumber: int
    objectMid: str
    project: ProjectWithoutDataset
    projectId: str
    status: IssueStatus
    type: "type"
    updatedAt: str


@deprecate()
class Asset(TypedDict):
    """
    A wrapper for Asset GraphQL object.
    """

    id: str
    consensusMark: float
    consensusMarkCompute: float
    consensusMarkPerCategory: Dict
    content: str
    contentJson: Dict
    contentJsonCompute: Dict
    createdAt: str
    duration: float
    durationCompute: float
    externalId: str
    honeypotMark: float
    honeypotMarkCompute: float
    inferenceMark: float
    inferenceMarkCompute: float
    isHoneypot: bool
    isToBeLabeledBy: bool
    issues: IssueWithoutAsset
    isUsedForConsensus: bool
    jsonContent: Dict
    jsonMetadata: Dict
    labels: LabelWithoutLabelOf
    latestLabel: LabelWithoutLabelOf
    latestLabelCompute: LabelWithoutLabelOf
    locks: Lock
    metadataCompute: Dict
    metadata: Dict
    numberOfValidLocks: int
    numberOfValidLocksCompute: int
    ocrMetadata: Dict
    priority: int
    project: ProjectWithoutDataset
    projectId: str
    projectIdCompute: str
    readPermissionsFromLabels: str
    skipped: bool
    status: Status
    statusCompute: Status
    thumbnail: str
    thumbnailCompute: str
    toBeLabeledBy: ProjectUser
    updatedAt: str


@deprecate()
class Label(LabelWithoutLabelOf):
    """
    A wrapper for Label GraphQL object.
    """

    labelOf: Asset


@deprecate()
class Project(ProjectWithoutDataset, total=False):
    """
    A wrapper for Project GraphQL object.
    """

    dataset: Asset


@deprecate()
class Notification(TypedDict):
    """
    A wrapper for Notification GraphQL object.
    """

    id: str
    createdAt: str
    hasBeenSeen: bool
    message: str
    status: NotificationStatus
    url: str
    userID: str


@deprecate()
class Plugin(TypedDict):
    """
    A wrapper for Plugin GraphQL object.
    """

    id: str
    name: str
    projectIds: List[str]
    createdAt: str
    updatedAt: str


@deprecate()
class ProjectVersion(TypedDict):
    """
    A wrapper for ProjectVersion GraphQL object.
    """

    id: str
    createdAt: str
    content: str
    name: str
    project: Project
    projectId: str


@deprecate()
class Issue(IssueWithoutAsset):
    """
    A wrapper for Issue GraphQL object.
    """

    asset: Asset


@deprecate()
class Comment(CommentWithoutIssue):
    """
    A wrapper for Comment GraphQL object.
    """

    issue: Issue


# pylint: enable:invalid-name



"""
This script defines object-relational mapping helpers to ease
the manipulation of Kili data structures.
"""
from dataclasses import dataclass

from typing_extensions import Literal


class DictClass(dict):
    """
    A python class that acts like dict and can be used to access its attributes as dict keys
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__dict__ = self


AnnotationFormat = Literal[
    "kili", "raw", "simple", "yolo_v4", "yolo_v5", "yolo_v7", "coco", "pascal_voc"
]


@dataclass
class AssetStatus:
    """
    List of asset status
    """

    Labeled = "LABELED"
    Ongoing = "ONGOING"
    Reviewed = "REVIEWED"
    ToReview = "TO_REVIEW"
    Todo = "TODO"


@dataclass
class JobMLTask:
    """
    List of ML Tasks
    """

    AssetAnnotation = "ASSET_ANNOTATION"
    Classification = "CLASSIFICATION"
    NamedEntitiesRecognition = "NAMED_ENTITIES_RECOGNITION"
    NamedEntitiesRelation = "NAMED_ENTITIES_RELATION"
    ObjectRelation = "OBJECT_RELATION"
    ObjectDetection = "OBJECT_DETECTION"
    PoseEstimation = "POSE_ESTIMATION"
    RegionDetection = "REGION_DETECTION"
    SpeechToText = "SPEECH_TO_TEXT"
    Transcription = "TRANSCRIPTION"


@dataclass
class JobTool:
    """
    List of tools
    """

    Marker = "marker"
    Polygon = "polygon"
    Polyline = "polyline"
    Range = "range"
    Rectangle = "rectangle"
    Semantic = "semantic"
    Vector = "vector"


def get_polygon(annotation):
    """
    Extracts a polygon from an annotation

    Args:
        annotation : Kili annotation
    """
    try:
        return annotation["boundingPoly"][0]["normalizedVertices"]
    except KeyError:
        return None


def get_category(annotation):
    """
    Extracts a category from an annotation

    Args:
        annotation : Kili annotation
    """
    try:
        return annotation["categories"][0]["name"]
    except KeyError:
        return None


def get_named_entity(annotation):
    """
    Extracts a named entity from an annotation

    Args:
        annotation : Kili annotation
    """
    try:
        return {
            "beginId": annotation["beginId"],
            "beginOffset": annotation["beginOffset"],
            "content": annotation["content"],
            "endId": annotation["endId"],
            "endOffset": annotation["endOffset"],
        }
    except KeyError:
        return None


def format_image_annotation(annotation):
    """
    Extracts a category, polygon, named entity from an annotation
    depending of the annotation type

    Args:
        annotation : Kili annotation
    """
    category = get_category(annotation)
    polygon = get_polygon(annotation)
    named_entity = get_named_entity(annotation)
    if polygon is not None:
        return (category, polygon)
    if named_entity is not None:
        return (category, named_entity)
    if category is not None:
        return category
    return None


class Label(DictClass):
    """
    Label class
    """

    jsonResponse = {}

    def json_response(self, _format: AnnotationFormat = "raw"):
        """
        Format a json response

        Args:
            _format: expected format
        """
        if "jsonResponse" not in self:
            raise KeyError(
                f'You did not fetch jsonResponse for label "{self["id"] if "id" in self else self}"'
            )
        if _format == "simple":
            job_names = self.jsonResponse.keys()
            if len(job_names) > 1:
                return {
                    "error": (
                        "Simple format is not adapted"
                        " when there is more than one job."
                        " Please choose another annotation format."
                    )
                }
            for job_name in job_names:
                job_response = self.jsonResponse[job_name]
                category = get_category(job_response)
                if category is not None:
                    return category
                if "annotations" not in job_response:
                    return None
                return [
                    format_image_annotation(annotation)
                    for annotation in job_response["annotations"]
                ]
            return None
        return self.jsonResponse


class Asset(DictClass):
    """
    Asset class
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)  # will add keys in *args and **kwargs to self
        if "labels" in self:
            labels = []
            for label in self["labels"]:
                labels.append(Label(label))
            self.labels = labels
        if "latestLabel" in self:
            latest_label = self["latestLabel"]
            if latest_label is not None:
                self.latestLabel = Label(latest_label)  # pylint: disable=invalid-name



"""
GraphQL Queries of Issues
"""

from typing import List, Optional

from typing_extensions import Literal

from kili.graphql import BaseQueryWhere, GraphQLQuery


class IssueWhere(BaseQueryWhere):
    """
    Tuple to be passed to the IssueQuery to restrict query
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        project_id: Optional[str] = None,
        asset_id: Optional[str] = None,
        asset_id_in: Optional[List[str]] = None,
        issue_type: Optional[Literal["QUESTION", "ISSUE"]] = None,
        status: Optional[Literal["OPEN", "SOLVED"]] = None,
    ):
        self.project_id = project_id
        self.asset_id = asset_id
        self.asset_id_in = asset_id_in
        self.issue_type = issue_type
        self.status = status
        super().__init__()

    def graphql_where_builder(self):
        """Build the GraphQL Where payload sent in the resolver from the SDK IssueWhere"""
        return {
            "project": {"id": self.project_id},
            "asset": {"id": self.asset_id},
            "assetIn": self.asset_id_in,
            "status": self.status,
            "type": self.issue_type,
        }


class IssueQuery(GraphQLQuery):
    """Issue query."""

    @staticmethod
    def query(fragment):
        """
        Return the GraphQL issues query
        """
        return f"""
        query issues($where: IssueWhere!, $first: PageSize!, $skip: Int!) {{
            data: issues(where: $where, first: $first, skip: $skip) {{
                {fragment}
            }}
        }}
        """

    COUNT_QUERY = """
    query countIssues($where: IssueWhere!) {
        data: countIssues(where: $where)
    }
    """



"""Develop Plugins for Kili"""

import logging
from typing import Dict, Optional

from kili.client import Kili
from kili.services.plugins.helpers import get_logger


class PluginCore:
    """
    Kili Plugin core class

    Args:
        kili: kili instance that plugins can make use of
        project_id: the project on which plugin is ran

    Implements:

        on_submit(self, label: Label, asset_id: str)
        on_review(self, label: Label, asset_id: str)

    # Warning : if using a custom init, be sure to call super().__init__()
    """

    logger: logging.Logger
    kili: Kili
    project_id: str

    def __init__(
        self, kili: Kili, project_id: str, logger: Optional[logging.Logger] = None
    ) -> None:
        self.kili = kili
        self.project_id = project_id
        if logger:
            self.logger = logger
        else:
            self.logger = get_logger()

    def on_submit(
        self,
        label: Dict,
        asset_id: str,
    ) -> None:
        """
        Handler for the submit action, triggered when a default label is submitted into Kili.

        Args:
            label: Label submitted to Kili: a dictionary containing the following fields:
                `id`, `labelType`, `numberOfAnnotations`, `authorId`, `modelName`, `jsonResponse`,
                `secondsToLabel`, `isSentBackToQueue`, `search` and some technical fields:
                `createdAt`, `updatedAt`, `version`, `isLatestReviewLabelForUser`,
                `isLatestLabelForUser`, `isLatestDefaultLabelForUser`, `readPermissionsFromProject`.
            asset_id: Id of the asset on which the label was submitted

        Example use:

            >>> def on_submit(self, label: Label, asset_id: str):
            >>>     json_response = label.get('jsonResponse')
            >>>     if label_is_respecting_business_rule(json_response):
            >>>         return
            >>>     else:
            >>>         self.kili.send_back_to_queue(asset_ids=[asset_id])

        """
        # pylint: disable=unused-argument
        self.logger.warning("Method not implemented. Define a custom on_submit on your plugin")
        pass  # pylint: disable=unnecessary-pass

    def on_review(
        self,
        label: Dict,
        asset_id: str,
    ) -> None:
        """
        Handler for the review action, triggered when a default label is reviewed on Kili

        Args:
            label: Label submitted to Kili: a dictionary containing the following fields:
                `id`, `labelType`, `numberOfAnnotations`, `authorId`, `modelName`, `jsonResponse`,
                `secondsToLabel`, `isSentBackToQueue`, `search` and `reviewedLabel` (dictionary
                that has a field `id` representing the id of the original label that was reviewed).
                It also contains some technical fields: `createdAt`, `updatedAt`, `version`,
                `isLatestReviewLabelForUser`, `isLatestLabelForUser`, `isLatestDefaultLabelForUser`,
                `readPermissionsFromProject`.
            asset_id: Id of the asset on which the label was submitted

        Example use:

            >>> def on_review(self, label: Label, asset_id: str):
            >>>     json_response = label.get('jsonResponse')
            >>>     if label_is_respecting_business_rule(json_response):
            >>>         return
            >>>     else:
            >>>         self.kili.send_back_to_queue(asset_ids=[asset_id])
        """
        # pylint: disable=unused-argument
        self.logger.warning("Method not implemented. Define a custom on_review on your plugin")
        pass  # pylint: disable=unnecessary-pass

    def on_custom_interface_click(
        self,
        label: Dict,
        asset_id: str,
    ) -> None:
        """
        Handler for the custom interface click action.

        **Warning**: This handler is in beta and is still in active development,
        so it should not be used.

        Args:
            label: label submitted to Kili
            asset_id: id of the asset on which the action is called
        """
        # pylint: disable=unused-argument
        self.logger.warning("Handler is in active development.")



"""
GraphQL Queries of ProjectVersions
"""

from kili.graphql import BaseQueryWhere, GraphQLQuery


class ProjectVersionWhere(BaseQueryWhere):
    """
    Tuple to be passed to the ProjectVersionQuery to restrict query
    """

    def __init__(
        self,
        project_id: str,
    ):
        self.project_id = project_id
        super().__init__()

    def graphql_where_builder(self):
        """Build the GraphQL Where payload sent in the resolver from the SDK ProjectVersionWhere"""
        return {
            "projectId": self.project_id,
        }


class ProjectVersionQuery(GraphQLQuery):
    """ProjectVersion query."""

    @staticmethod
    def query(fragment):
        """
        Return the GraphQL projects query
        """
        return f"""
        query projectVersions($where: ProjectVersionWhere!, $first: PageSize!, $skip: Int!) {{
            data: projectVersions(where: $where, first: $first, skip: $skip) {{
                {fragment}
            }}
        }}
        """

    COUNT_QUERY = """
    query countProjectVersions($where: ProjectVersionWhere!) {
        data: countProjectVersions(where: $where)
    }
    """



"""Data connection queries."""

from typing import Dict, Generator, Iterable, List, Optional, overload

from typeguard import typechecked
from typing_extensions import Literal

from kili.authentication import KiliAuth
from kili.graphql import QueryOptions
from kili.graphql.operations.data_connection.queries import (
    DataConnectionsQuery,
    DataConnectionsWhere,
)
from kili.helpers import disable_tqdm_if_as_generator
from kili.utils.logcontext import for_all_methods, log_call

from ... import services


@for_all_methods(log_call, exclude=["__init__"])
# pylint: disable=too-few-public-methods
class QueriesDataConnection:
    """
    Set of cloud storage connection queries
    """

    # pylint: disable=too-many-arguments,dangerous-default-value

    def __init__(self, auth: KiliAuth):
        """Initialize the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    @overload
    def cloud_storage_connections(
        self,
        cloud_storage_connection_id: Optional[str] = None,
        cloud_storage_integration_id: Optional[str] = None,
        project_id: Optional[str] = None,
        fields: List[str] = [
            "id",
            "lastChecked",
            "numberOfAssets",
            "selectedFolders",
            "projectId",
        ],
        first: Optional[int] = None,
        skip: int = 0,
        disable_tqdm: bool = False,
        *,
        as_generator: Literal[True],
    ) -> Generator[Dict, None, None]:
        ...

    @overload
    def cloud_storage_connections(
        self,
        cloud_storage_connection_id: Optional[str] = None,
        cloud_storage_integration_id: Optional[str] = None,
        project_id: Optional[str] = None,
        fields: List[str] = [
            "id",
            "lastChecked",
            "numberOfAssets",
            "selectedFolders",
            "projectId",
        ],
        first: Optional[int] = None,
        skip: int = 0,
        disable_tqdm: bool = False,
        *,
        as_generator: Literal[False] = False,
    ) -> List[Dict]:
        ...

    @typechecked
    def cloud_storage_connections(
        self,
        cloud_storage_connection_id: Optional[str] = None,
        cloud_storage_integration_id: Optional[str] = None,
        project_id: Optional[str] = None,
        fields: List[str] = [
            "id",
            "lastChecked",
            "numberOfAssets",
            "selectedFolders",
            "projectId",
        ],
        first: Optional[int] = None,
        skip: int = 0,
        disable_tqdm: bool = False,
        *,
        as_generator: bool = False,
    ) -> Iterable[Dict]:
        # pylint: disable=line-too-long
        """Get a generator or a list of cloud storage connections that match a set of criteria.

        Args:
            cloud_storage_connection_id: ID of the cloud storage connection.
            cloud_storage_integration_id: ID of the cloud storage integration.
            project_id: ID of the project.
            fields: All the fields to request among the possible fields for the cloud storage connections.
                See [the documentation](https://docs.kili-technology.com/reference/graphql-api#dataconnection) for all possible fields.
            first: Maximum number of cloud storage connections to return.
            skip: Number of skipped cloud storage connections.
            disable_tqdm: If `True`, the progress bar will be disabled.
            as_generator: If `True`, a generator on the cloud storage connections is returned.

        Returns:
            A list or a generator of the cloud storage connections that match the criteria.

        Examples:
            >>> kili.cloud_storage_connections(project_id="789465123")
            [{'id': '123456789', 'lastChecked': '2023-02-21T14:49:35.606Z', 'numberOfAssets': 42, 'selectedFolders': ['folder1', 'folder2'], 'projectId': '789465123'}]
        """
        if (
            cloud_storage_connection_id is None
            and cloud_storage_integration_id is None
            and project_id is None
        ):
            raise ValueError(
                "At least one of cloud_storage_connection_id, cloud_storage_integration_id or"
                " project_id must be specified"
            )

        # call dataConnection resolver
        if cloud_storage_connection_id is not None:
            data_connection = services.get_data_connection(
                self.auth, cloud_storage_connection_id, fields
            )
            data_connection_list = [data_connection]
            if as_generator:
                return iter(data_connection_list)
            return data_connection_list

        # call dataConnections resolver
        where = DataConnectionsWhere(
            project_id=project_id, data_integration_id=cloud_storage_integration_id
        )
        disable_tqdm = disable_tqdm_if_as_generator(as_generator, disable_tqdm)
        options = QueryOptions(disable_tqdm, first, skip)
        data_connections_gen = DataConnectionsQuery(self.auth.client, self.auth.ssl_verify)(
            where, fields, options
        )

        if as_generator:
            return data_connections_gen
        return list(data_connections_gen)



"""
Helpers for GraphQL Queries and Mutations
"""

import base64
import functools
import glob
import mimetypes
import os
import re
import warnings
from json import dumps, loads
from typing import Any, Callable, Dict, List, Optional, Type, TypeVar, Union

import pyparsing as pp
import requests
import tenacity
from typing_extensions import get_args, get_origin

from kili.constants import mime_extensions_for_IV2

T = TypeVar("T")


def format_result(
    name: str, result: dict, _object: Optional[Type[T]], ssl_verify: Union[bool, str]
) -> T:
    """
    Formats the result of the GraphQL queries.
    Args:
        name: name of the field to extract, usually data
        result: query result to parse
        _object: returned type
        ssl_verify: requests' ssl_verify parameter
    """
    formatted_json = format_json(result[name], ssl_verify)
    if _object is None:
        return formatted_json  # type:ignore X
    if isinstance(formatted_json, list):
        if get_origin(_object) is list:
            obj = get_args(_object)[0]
            return [obj(element) for element in formatted_json]  # type:ignore
        # the legacy "orm" objects fall into this category.
        return [_object(element) for element in formatted_json]  # type:ignore

    return _object(formatted_json)  # type:ignore


def content_escape(content: str) -> str:
    """
    Escapes the content

    Args:
        content (str): string to escape

    Returns:
        str: escaped string
    """
    return content.replace("\\", "\\\\").replace("\n", "\\n").replace('"', '\\"')


def get_data_type(path: str):
    """
    Get the data type, either image/png or application/pdf

    Args:
        path: path of the file
    """
    mime_type, _ = mimetypes.guess_type(path.lower())
    return mime_type if mime_type else ""


def encode_base64(path):
    """
    Encode a file in base 64

    Args:
        path: path of the file
    """
    data_type = get_data_type(path)
    with open(path, "rb") as image_file:
        return f"data:{data_type};base64," + base64.b64encode(image_file.read()).decode("ascii")


def is_url(path):
    """
    Check if the path is a url or something else

    Args:
        path: path of the file
    """
    return isinstance(path, str) and re.match(r"^(http://|https://)", path.lower())


def format_json_dict(result: dict, ssl_verify: Union[bool, str]) -> Dict:
    """
    Formats the dict part of a json return by a GraphQL query into a python object

    Args:
        result: result of a GraphQL query
    """
    for key, value in result.items():
        if key in ["jsonInterface", "jsonMetadata", "jsonResponse"]:
            if (value == "" or value is None) and not (is_url(value) and key == "jsonInterface"):
                result[key] = {}
            elif isinstance(value, str):
                try:
                    if is_url(value):
                        result[key] = requests.get(value, verify=ssl_verify, timeout=30).json()
                    else:
                        result[key] = loads(value)
                except Exception as exception:
                    raise ValueError(
                        "Json Metadata / json response / json interface should be valid jsons"
                    ) from exception
        else:
            result[key] = format_json(value, ssl_verify)
    return result


D = TypeVar("D")


def format_json(
    result: Union[None, list, dict, D], ssl_verify: Union[bool, str]
) -> Union[None, list, dict, D]:
    """
    Formats the json return by a GraphQL query into a python object

    Args:
        result: result of a GraphQL query
    """
    if result is None:
        return result
    if isinstance(result, list):
        return [format_json(elem, ssl_verify) for elem in result]
    if isinstance(result, dict):
        return format_json_dict(result, ssl_verify)
    return result


def deprecate(
    msg: Optional[str] = None,
    removed_in: Optional[str] = None,
    _type=DeprecationWarning,
):
    """
    Decorator factory that tag a deprecated function.
    - To deprecated the whole function, you can give a message at the decorator level.
    - For more sharp condition on the warning message, integrate this warning inside the function
    but still tag the function with this decorator, without giving a message argument

    Args:
        msg: string message that will be displayed whenever the function is called
        removed_in: string version in the format "Major.Minor"
            in which the deprecation element has to be removed
        type: DeprecationWarning by default
    """

    def decorator(func):
        if removed_in:
            if len(removed_in.split(".")) != 2:
                raise ValueError(
                    f'"removed_in" argument in deprecate wrapper of the function {func.__name__}'
                    'should have the format "Major.Minor"'
                )
            func.removed_in = removed_in

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if msg:
                warnings.warn(msg, _type, stacklevel=2)
            return func(*args, **kwargs)

        return wrapper

    return decorator


def format_metadata(metadata):
    """
    Formats metadata

    Args:
        metadata: a python object
    """
    if metadata is None:
        return metadata
    if isinstance(metadata, str):
        return metadata
    if isinstance(metadata, (dict, list)):
        return dumps(metadata)
    raise TypeError(
        f"Metadata {metadata} of type {type(metadata)} must either be None,"
        " a string a list or a dict."
    )


def convert_to_list_of_none(array, length):
    """
    Turns a value in a list of length length

    Args:
        array
        length
    """
    if isinstance(array, list):
        if len(array) != length:
            raise ValueError(f"array should have length {length}")
        return array
    return [None] * length


def is_none_or_empty(_object):
    """
    Tests if an object is none or empty

    Args:
        object: a python object
    """
    object_is_empty = isinstance(_object, list) and len(_object) == 0
    return _object is None or object_is_empty


def list_is_not_none_else_none(_object):
    """
    Formats an object as a singleton if not none

    Args:
        object: a python object
    """
    return [_object] if _object is not None else None


def validate_category_search_query(query):
    """Validate the category search query
    Args:
        query: the query to parse

    Raises:
        ValueError: if `query` is invalid
    """
    operator = pp.oneOf(">= <= > < ==")
    number = pp.pyparsing_common.number()
    dot = "."
    word = pp.Word(pp.alphas, pp.alphanums + "_-*")
    identifier = word + dot + word + dot + "count"
    condition = identifier + operator + number

    expr = pp.infixNotation(
        condition,
        [
            (
                "AND",
                2,
                pp.opAssoc.LEFT,
            ),
            (
                "OR",
                2,
                pp.opAssoc.LEFT,
            ),
        ],
    )
    try:
        expr.parseString(query, parseAll=True)
    except pp.ParseException as error:
        raise ValueError(f"Invalid category search query: {query}") from error


def get_file_paths_to_upload(
    files: List[str], file_check_function: Optional[Callable] = None, verbose: bool = False
) -> List[str]:
    """Get a list of paths for the files to upload given a list of files or folder paths.

    Args:
        files: a list path that can either be file paths, folder paths or unexisting paths
        file_check_function: function to check files. Must use argument path and return a bool
    Returns:
        a list of the paths of the files to upload, compatible with the project type.
    """
    file_check_function = file_check_function or (lambda x: True)
    file_paths = []
    for item in files:
        if os.path.isfile(item):
            file_paths.append(item)
        elif os.path.isdir(item):
            folder_path = os.path.join(item, "")
            file_paths.extend(
                [sub_item for sub_item in glob.glob(folder_path + "*") if os.path.isfile(sub_item)]
            )
        else:
            file_paths.extend(
                [sub_item for sub_item in glob.glob(item) if os.path.isfile(sub_item)]
            )
    file_paths_to_upload = [path for path in file_paths if file_check_function(path)]
    if len(file_paths_to_upload) == 0:
        raise ValueError(
            "No files to upload. Check that the paths exist and that the file type is correct"
        )
    if verbose:
        for path in file_paths:
            if path not in file_paths_to_upload:
                print(f"{path:30} SKIPPED")
        if len(file_paths_to_upload) != len(file_paths):
            print("Paths skipped either do not exist or point towards an incorrect file")
    file_paths_to_upload.sort()
    return file_paths_to_upload


def file_check_function_from_input_type(input_type: str):
    """
    Returns check_file_mime_type function with input_type and verbose as preset argument
    """

    def output_function(path: str):
        return check_file_mime_type(path, input_type, raise_error=False)

    return output_function


def check_file_mime_type(path: str, input_type: str, raise_error=True) -> bool:
    """
    Returns true if the mime type of the file corresponds to the allowed mime types of the project
    """

    mime_type = get_data_type(path.lower())

    if not (mime_extensions_for_IV2[input_type] and mime_type):
        return False

    correct_mime_type = mime_type in mime_extensions_for_IV2[input_type]
    if not correct_mime_type and raise_error:
        raise ValueError(
            f"File mime type for {path} is {mime_type} and does not correspond "
            "to the type of the project. "
            f"File mime type should be one of {mime_extensions_for_IV2[input_type]}"
        )
    return correct_mime_type


def disable_tqdm_if_as_generator(as_generator: bool, disable_tqdm: bool):
    """Disable tqdm in user-facing queries method if the return type is asked as a generator."""
    if as_generator and not disable_tqdm:
        disable_tqdm = True
        warnings.warn(
            (
                "tqdm has been forced disabled because its behavior is not compatible with the"
                " generator return type"
            ),
            stacklevel=2,
        )
    return disable_tqdm


class RetryLongWaitWarner:  # pylint: disable=too-few-public-methods
    """Class that warns when retry takes too long."""

    def __init__(
        self,
        warn_message: str,
        logger_func: Callable,
        warn_after: float = 10,
    ) -> None:
        """Class that warns when retry takes too long.

        Args:
            warn_message: custom warning message. If not provided, a default message is used.
            logger_func: function to log the message (print, warning, logger.warning, etc.)
            warn_after: time in seconds after which the warning is raised.
        """
        self.warn_message = warn_message
        self.logger_func = logger_func
        self.warn_after = warn_after

        self.warned = False

    def __call__(self, retry_state: tenacity.RetryCallState):
        if not self.warned and float(retry_state.outcome_timestamp or 0) > self.warn_after:
            self.logger_func(self.warn_message)
            self.warned = True


def is_empty_list_with_warning(method_name: str, argument_name: str, argument_value: Any) -> bool:
    """
    Check if an input list argument is empty and warn the user if it is

    Returns True if the list is empty, False otherwise
    """
    if isinstance(argument_value, List) and len(argument_value) == 0:
        warnings.warn(
            (
                f"Method '{method_name}' did nothing because the following argument"
                f" is empty: {argument_name}."
            ),
            stacklevel=5,
        )
        return True
    return False



"""
GraphQL Client
"""

import json
import random
import string
import threading
import time
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, Optional, Union
from urllib.parse import urlparse

import graphql
import requests
import websocket
from gql import Client, gql
from gql.transport import exceptions
from gql.transport.requests import RequestsHTTPTransport
from graphql import DocumentNode, print_schema

from kili import __version__
from kili.exceptions import GraphQLError
from kili.graphql.clientnames import GraphQLClientName
from kili.utils.logcontext import LogContext


class GraphQLClient:
    """
    GraphQL client
    """

    def __init__(
        self,
        endpoint: str,
        api_key: str,
        client_name: GraphQLClientName,
        ssl_verify: Union[bool, str],
    ) -> None:
        self.endpoint = endpoint
        self.api_key = api_key
        self.client_name = client_name
        self.ssl_verify = ssl_verify

        self.ws_endpoint = self.endpoint.replace("http", "ws")

        self._gql_transport = RequestsHTTPTransport(
            url=endpoint,
            headers=self._get_headers(),
            cookies=None,
            auth=None,
            use_json=True,
            timeout=30,
            verify=ssl_verify,
            retries=20,
            method="POST",
        )

        self._gql_client = self._initizalize_graphql_client()

    def _get_headers(self) -> Dict[str, str]:
        """
        Get the headers
        """
        return {
            "Authorization": f"X-API-Key: {self.api_key}",
            "Accept": "application/json",
            "Content-Type": "application/json",
            "apollographql-client-name": self.client_name.value,
            "apollographql-client-version": __version__,
        }

    def _initizalize_graphql_client(self) -> Client:
        """
        Initialize the GraphQL client
        """
        graphql_schema_path = self._get_graphql_schema_path()

        # In some cases (local development), we cannot get the kili version from the backend
        # and therefore we cannot determine the schema version, so we don't cache the schema
        if graphql_schema_path is None:
            return Client(transport=self._gql_transport, fetch_schema_from_transport=True)

        # If the schema is not in the cache, we fetch it from the backend and cache it
        if not (graphql_schema_path.is_file() and graphql_schema_path.stat().st_size > 0):
            self._purge_graphql_schema_cache_dir()
            self._cache_graphql_schema(graphql_schema_path)

        return Client(
            schema=graphql_schema_path.read_text(encoding="utf-8"),
            transport=self._gql_transport,
        )

    def _cache_graphql_schema(self, graphql_schema_path: Path) -> None:
        """
        Cache the graphql schema on disk.
        """
        with Client(transport=self._gql_transport, fetch_schema_from_transport=True) as session:
            schema_str = print_schema(session.client.schema)  # type: ignore

        graphql_schema_path.parent.mkdir(parents=True, exist_ok=True)

        with graphql_schema_path.open("w", encoding="utf-8") as file:
            file.write(schema_str)

    @property
    def graphql_schema_cache_dir(self) -> Path:
        """
        Get the path of the GraphQL schema cache directory
        """
        return Path.home() / ".cache" / "kili" / "graphql"

    def _purge_graphql_schema_cache_dir(self) -> None:
        """
        Purge the schema cache directory
        """
        for file in self.graphql_schema_cache_dir.glob("*.graphql"):
            file.unlink()

    def _get_graphql_schema_path(self) -> Optional[Path]:
        """
        Get the path of the GraphQL schema

        Will return None if we cannot get the schema version.
        """
        endpoint_netloc = urlparse(self.endpoint).netloc
        version = self._get_kili_app_version()
        if version is None:
            return None
        filename = f"{endpoint_netloc}_{version}.graphql"
        return self.graphql_schema_cache_dir / filename

    def _get_kili_app_version(self) -> Optional[str]:
        """
        Get the version of the Kili app server

        Returns None if the version cannot be retrieved.
        """
        url = self.endpoint.replace("/graphql", "/version")
        response = requests.get(url, verify=self.ssl_verify, timeout=30)
        if response.status_code == 200 and '"version":' in response.text:
            response_json = response.json()
            version = response_json["version"]
            return version
        return None

    def execute(
        self, query: Union[str, DocumentNode], variables: Optional[Dict] = None, **kwargs
    ) -> Dict[str, Any]:
        """
        Execute a query

        Args:
            query: the GraphQL query
            variables: the payload of the query
            kwargs: additional arguments to pass to the GraphQL client
        """

        def _execute(
            document: DocumentNode, variables: Optional[Dict] = None, **kwargs
        ) -> Dict[str, Any]:
            try:
                result = self._gql_client.execute(
                    document=document,
                    variable_values=variables,
                    extra_args={
                        "headers": {
                            **self._gql_transport.headers,  # type: ignore
                            **LogContext(),
                        }
                    },
                    **kwargs,
                )
            except (exceptions.TransportQueryError, graphql.GraphQLError) as err:
                if isinstance(err, exceptions.TransportQueryError):
                    raise GraphQLError(error=err.errors) from err
                if isinstance(err, graphql.GraphQLError):
                    raise GraphQLError(error=err.message) from err
            return result  # type: ignore

        document = query if isinstance(query, DocumentNode) else gql(query)

        try:
            ret = _execute(document, variables, **kwargs)
        except GraphQLError as err:
            # if error is due do parsing or local validation of the query (graphql.GraphQLError)
            # we refresh the schema and retry once
            if isinstance(err.__cause__, graphql.GraphQLError):
                self._purge_graphql_schema_cache_dir()
                self._gql_client = self._initizalize_graphql_client()
                ret = _execute(document, variables, **kwargs)  # if it fails again, we crash here
            else:
                raise err
        return ret


GQL_WS_SUBPROTOCOL = "graphql-ws"


class SubscriptionGraphQLClient:
    """
    A simple GraphQL client that works over Websocket as the transport
    protocol, instead of HTTP.
    This follows the Apollo protocol.
    https://github.com/apollographql/subscriptions-transport-ws/blob/master/PROTOCOL.md
    """

    # pylint: disable=too-many-instance-attributes, too-many-arguments

    def __init__(self, url):
        self.ws_url = url
        self._id = None
        self._paused = False
        self._connect()
        self._subscription_running = False
        self._st_id = None
        self.failed_connection_attempts = 0

    def _connect(self):
        """
        Handles the connection
        """
        self._conn = websocket.create_connection(
            self.ws_url, on_message=self._on_message, subprotocols=[GQL_WS_SUBPROTOCOL]
        )
        self._created_at = datetime.now()
        self._conn.on_message = self._on_message  # type: ignore

    def _reconnect(self):
        """
        Handles the reconnection
        """
        self._connect()
        self._subscription_running = True
        dt_string = datetime.now().strftime(r"%d/%m/%Y %H:%M:%S")
        print(f"{dt_string} reconnected")
        self.failed_connection_attempts = 0

    def _on_message(self, message):
        """
        Handles messages

        Args:
            message : the message
        """
        data = json.loads(message)
        # skip keepalive messages
        if data["type"] != "ka":
            print(message)

    def _conn_init(self, headers=None, authorization=None):
        """
        Initializes the websocket connection

        Args:
            headers : Headers are necessary for Kili API v1
            authorization : Headers are necessary for Kili API v2
        """
        payload = {
            "type": "connection_init",
            "payload": {"headers": headers, "Authorization": authorization},
        }
        self._conn.send(json.dumps(payload).encode("utf-8"))
        self._conn.recv()

    def _start(self, payload):
        """
        Handles start

        Args:
            payload
        """
        _id = gen_id()
        frame = {"id": _id, "type": "start", "payload": payload}
        self._conn.send(json.dumps(frame).encode("utf-8"))
        return _id

    def _stop(self, _id):
        """
        Handles stop

        Args:
        - _id: connection id
        """
        payload = {"id": _id, "type": "stop"}
        self._conn.send(json.dumps(payload).encode("utf-8"))
        return self._conn.recv()

    def query(self, query, variables=None, headers=None):
        """
        Sends a query

        Args:
            query
            variables
            headers
        """
        self._conn_init(headers)
        payload = {"headers": headers, "query": query, "variables": variables}
        _id = self._start(payload)
        res = self._conn.recv()
        self._stop(_id)
        return res

    def prepare_subscribe(self, query, variables, headers, callback, authorization):
        """
        Prepares a subscription

        Args:
            query
            variables
            headers
            callback: function executed after the subscription
            authorization: authorization header
        """
        self._conn_init(headers, authorization)
        payload = {"headers": headers, "query": query, "variables": variables}
        _cc = self._on_message if not callback else callback
        _id = self._start(payload)
        self._id = _id
        return _cc, _id

    def subscribe(self, query, variables=None, headers=None, callback=None, authorization=None):
        """
        Subscribes

        Args:
            query
            variables
            headers
            callback: function executed after the subscription
            authorization: authorization header
        """
        _cc, _id = self.prepare_subscribe(query, variables, headers, callback, authorization)

        def subs(_cc, _id):
            max_reconnections = 10
            self._subscription_running = True
            while (
                self._subscription_running and self.failed_connection_attempts < max_reconnections
            ):
                try:
                    response = json.loads(self._conn.recv())
                    if response["type"] == "error" or response["type"] == "complete":
                        print(response)
                        self._stop_subscribe(_id)
                        break
                    if response["type"] != "ka" and not self._paused:
                        _cc(_id, response)  # type:ignore
                    time.sleep(1)
                except (
                    websocket._exceptions.WebSocketConnectionClosedException  # pylint: disable=protected-access
                ) as error:
                    self.failed_connection_attempts += 1
                    dt_string = datetime.now().strftime("%d/%m/%Y %H:%M:%S")
                    error_message = str(error)
                    print(f"{dt_string} Connection closed error : {error_message}")
                    print(
                        "Will try to reconnect"
                        f" {max_reconnections - self.failed_connection_attempts} times..."
                    )
                    self._reconnect()
                    _cc, _id = self.prepare_subscribe(
                        query, variables, headers, callback, authorization
                    )
                    continue
            print(f"Did not reconnect successfully after {max_reconnections} attempts")

        self._st_id = threading.Thread(target=subs, args=(_cc, _id))
        self._st_id.start()
        return _id

    def _stop_subscribe(self, _id):
        self._subscription_running = False
        self._stop(_id)

    def close(self):
        """
        Handles close
        """
        self._conn.close()

    def pause(self):
        """
        Handles pause
        """
        self._paused = True

    def unpause(self):
        """
        Handles unpause
        """
        self._paused = False

    def get_lifetime(self):
        """
        Return the lifetime
        """
        return (datetime.now() - self._created_at).seconds

    def reset_timeout(self):
        """
        Resets the timeout
        """
        self._reconnect()


def gen_id(size=6, chars=string.ascii_letters + string.digits):
    """
    Generate random alphanumeric id

    Args:
        size: length of the id
        chars: chars used to generate the id
    """
    return "".join(random.choice(chars) for _ in range(size))



"""
Utils
"""
import functools
import time
from time import sleep
from typing import Any, Callable, Dict, Iterator, List, Optional

from kili.constants import MUTATION_BATCH_SIZE, THROTTLING_DELAY
from kili.exceptions import GraphQLError

# pylint: disable=too-many-arguments


class BatchIteratorBuilder:
    """Generate an paginated iterator from a list
    Args:
        iterable: a list to paginate.
        batch_size: the size of the batches to produce
    """

    def __init__(self, iterable: List, batch_size=MUTATION_BATCH_SIZE) -> None:
        self.iterable = iterable
        self.batch_size = batch_size
        self.nb_batches = (len(iterable) - 1) // batch_size + 1

    def __len__(self):
        return self.nb_batches

    def __next__(self):
        next_batch = self.iterable[: self.batch_size]
        self.iterable = self.iterable[self.batch_size :]
        if len(next_batch) > 0:
            return next_batch
        raise StopIteration

    def __iter__(self):
        return self


def batch_object_builder(
    properties_to_batch: Dict[str, Optional[List[Any]]],
    batch_size: int = MUTATION_BATCH_SIZE,
) -> Iterator[Dict[str, Optional[List[Any]]]]:
    """Generate a paginated iterator for several variables
    Args:
        properties_to_batch: a dictionnary of properties to be batched.
        batch_size: the size of the batches to produce
    """
    if len(list(filter(None, properties_to_batch.values()))) == 0:
        yield properties_to_batch
        return
    number_of_objects = len([v for v in properties_to_batch.values() if v is not None][0])
    number_of_batches = len(range(0, number_of_objects, batch_size))
    batched_properties = {
        k: (
            BatchIteratorBuilder(iterable=v, batch_size=batch_size)
            if v is not None
            else (item for item in [v] * number_of_batches)
        )
        for k, v in properties_to_batch.items()
    }
    batch_object_iterator = (
        dict(zip(batched_properties, t)) for t in zip(*batched_properties.values())
    )
    for batch in batch_object_iterator:
        yield batch  # type: ignore


def api_throttle(func):
    """
    Define a decorator that throttle a function call to meet the API limitation
    """

    @functools.wraps(func)
    def throttled_wrapper(*args, **kwargs):
        call_start = time.time()
        result = func(*args, **kwargs)
        call_duration = time.time() - call_start
        if call_duration < THROTTLING_DELAY:
            time.sleep(THROTTLING_DELAY - call_duration)
        return result

    return throttled_wrapper


def _mutate_from_paginated_call(
    self,
    properties_to_batch: Dict[str, Optional[List[Any]]],
    generate_variables: Callable,
    request: str,
    batch_size: int = MUTATION_BATCH_SIZE,
    last_batch_callback: Optional[Callable] = None,
):
    """Run a mutation by making paginated calls
    Args:
        properties_to_batch: a dictionnary of properties to be batched.
            constants across batch are defined in the generate_variables function
        generate_variables: function that takes batched properties and return
            a graphQL payload for request for this batch
        request: the GraphQL request to call,
        batch_size: the size of the batches to produce
        last_batch_callback: a function that takes the last batch and the result of
            this method as arguments
    Example:
        '''
        properties_to_batch={prop1: [0,1], prop2: ['a', 'b']}
        def generate_variables(batched_properties):
            return {
                graphQL_prop1: batched_properties['prop1']
                graphQL_prop2: batched_properties['prop2']
            }
        _mutate_from_paginated_call(
                properties_to_batch=properties_to_batch,
                generate_variables=generate_variables
                request= APPEND_MANY_TO_DATASET
        )
        '''
    """
    results = []
    batch = None
    for batch_number, batch in enumerate(batch_object_builder(properties_to_batch, batch_size)):
        payload = generate_variables(batch)
        try:
            result = api_throttle(self.auth.client.execute)(request, payload)
        except GraphQLError as err:
            raise GraphQLError(error=err.error, batch_number=batch_number) from err
        results.append(result)

    sleep(1)  # wait for the backend to process the mutations
    if batch and results and last_batch_callback:
        last_batch_callback(batch, results)
    return results



"""Tools for export with videos."""

import os
import shutil
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

import ffmpeg

from kili.utils.tempfile import TemporaryDirectory


class FFmpegError(Exception):
    """Errors related to ffmpeg."""


def get_video_dimensions(file_path: Union[Path, str]) -> Tuple:
    """
    Get a video width and height
    """
    assert Path(file_path).is_file(), f"File {file_path} does not exist"
    probe = ffmpeg.probe(str(file_path))
    video_info = next(s for s in probe["streams"] if s["codec_type"] == "video")
    width = video_info["width"]
    height = video_info["height"]
    return width, height


def cut_video(
    video_path: Path, asset: Dict, leading_zeros: int, output_dir: Optional[Path]
) -> List[Path]:
    """Download and cut video into frames."""
    output_dir = output_dir or video_path.parent
    output_dir.mkdir(parents=True, exist_ok=True)

    if (
        "jsonMetadata" in asset
        and "processingParameters" in asset["jsonMetadata"]
        and "framesPlayedPerSecond" in asset["jsonMetadata"]["processingParameters"]
    ):
        metadata = asset["jsonMetadata"]
        final_framerate = metadata["processingParameters"]["framesPlayedPerSecond"]
    else:
        try:
            probe = ffmpeg.probe(str(video_path))
            video_info = next(s for s in probe["streams"] if s["codec_type"] == "video")
            # avg_frame_rate is total # of frames / total duration
            # r_frame_rate is the lowest framerate with which all timestamps can be represented
            # accurately (it is the least common multiple of all framerates in the stream)
            frame_rate_string = video_info["r_frame_rate"].split("/")
            final_framerate = int(frame_rate_string[0]) / int(frame_rate_string[1])
        except ffmpeg.Error as error:
            raise FFmpegError(f"ffmpeg error for asset {video_path.name}: {str(error)}") from error

    output_frames = []
    try:
        with TemporaryDirectory() as temp_dir:
            ffmpeg.input(str(video_path)).filter("fps", fps=final_framerate, round="up").output(
                os.path.join(str(temp_dir), "%d.jpg"), start_number=0
            ).run(capture_stdout=True, capture_stderr=True)
            for file in os.listdir(temp_dir):
                idx = int(Path(file).stem)
                new_filename = f'{asset["externalId"]}_{str(idx+1).zfill(leading_zeros)}.jpg'
                shutil.move(str(temp_dir / file), str(output_dir / new_filename))
                output_frames.append(output_dir / new_filename)
    except ffmpeg.Error as error:
        raise FFmpegError(f"ffmpeg error for asset {video_path.name}: {str(error)}") from error

    return sorted(output_frames)



"""
Asset mutations
"""

import warnings
from typing import Any, Dict, List, Optional, Union

from tenacity import retry
from tenacity.retry import retry_if_exception_type
from tenacity.wait import wait_exponential
from typeguard import typechecked

from kili.authentication import KiliAuth
from kili.exceptions import MissingArgumentError
from kili.graphql import QueryOptions
from kili.graphql.operations.asset.queries import AssetQuery, AssetWhere
from kili.helpers import format_result
from kili.mutations.asset.helpers import process_update_properties_in_assets_parameters
from kili.mutations.asset.queries import (
    GQL_ADD_ALL_LABELED_ASSETS_TO_REVIEW,
    GQL_DELETE_MANY_FROM_DATASET,
    GQL_SEND_BACK_ASSETS_TO_QUEUE,
    GQL_UPDATE_PROPERTIES_IN_ASSETS,
)
from kili.orm import Asset
from kili.services.asset_import import import_assets
from kili.utils.logcontext import for_all_methods, log_call
from kili.utils.pagination import _mutate_from_paginated_call

from ...helpers import is_empty_list_with_warning
from ..exceptions import MutationError
from .helpers import get_asset_ids_or_throw_error


@for_all_methods(log_call, exclude=["__init__"])
class MutationsAsset:
    """
    Set of Asset mutations
    """

    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth: KiliAuth):
        """Initialize the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    @typechecked
    def append_many_to_dataset(
        self,
        project_id: str,
        content_array: Optional[List[str]] = None,
        external_id_array: Optional[List[str]] = None,
        id_array: Optional[List[str]] = None,
        is_honeypot_array: Optional[List[bool]] = None,
        status_array: Optional[List[str]] = None,
        json_content_array: Optional[List[List[Union[dict, str]]]] = None,
        json_metadata_array: Optional[List[dict]] = None,
        disable_tqdm: bool = False,
        wait_until_availability: bool = True,
    ) -> Optional[Dict[str, str]]:
        # pylint: disable=line-too-long
        """Append assets to a project.

        Args:
            project_id: Identifier of the project
            content_array: List of elements added to the assets of the project
                Must not be None except if you provide json_content_array.

                - For a `TEXT` project, the content can be either raw text, or URLs to TEXT assets.
                - For an `IMAGE` / `PDF` project, the content can be either URLs or paths to existing
                    images/pdf on your computer.
                - For a VIDEO project, the content can be either URLs pointing to videos hosted on a web server or paths to
                existing video files on your computer. If you want to import video from frames, look at the json_content
                section below.
                - For an `VIDEO_LEGACY` project, the content can be only be URLs
            external_id_array: List of external ids given to identify the assets.
                If None, random identifiers are created.
            is_honeypot_array:  Whether to use the asset for honeypot
            status_array: By default, all imported assets are set to `TODO`. Other options:
                `ONGOING`, `LABELED`, `REVIEWED`.
            json_content_array: Useful for `VIDEO` or `TEXT` projects only.

                - For `VIDEO` projects, each element is a sequence of frames, i.e. a
                    list of URLs to images or a list of paths to images.
                - For `TEXT` projects, each element is a json_content dict,
                    formatted according to documentation [on how to import
                rich-text assets](https://github.com/kili-technology/kili-python-sdk/blob/master/recipes/import_text_assets.ipynb)
            json_metadata_array: The metadata given to each asset should be stored in a json like dict with keys.

                - Add metadata visible on the asset with the following keys: `imageUrl`, `text`, `url`.
                    Example for one asset: `json_metadata_array = [{'imageUrl': '','text': '','url': ''}]`.
                - For VIDEO projects (and not VIDEO_LEGACY), you can specify a value with key 'processingParameters' to specify the sampling rate (default: 30).
                    Example for one asset: `json_metadata_array = [{'processingParameters': {'framesPlayedPerSecond': 10}}]`.
            disable_tqdm: If `True`, the progress bar will be disabled
            wait_until_availability: If `True`, the function will return once the assets are fully imported in Kili.
                If `False`, the function will return faster but the assets might not be fully processed by the server.

        Returns:
            A result object which indicates if the mutation was successful, or an error message.

        Examples:
            >>> kili.append_many_to_dataset(
                    project_id=project_id,
                    content_array=['https://upload.wikimedia.org/wikipedia/en/7/7d/Lenna_%28test_image%29.png'])

        !!! example "Recipe"
            - For more detailed examples on how to import assets,
                see [the recipe](https://docs.kili-technology.com/recipes/importing-data).
            - For more detailed examples on how to import text assets,
                see [the recipe](https://github.com/kili-technology/kili-python-sdk/blob/master/recipes/import_text_assets.ipynb).
        """
        if is_empty_list_with_warning(
            "append_many_to_dataset", "content_array", content_array
        ) or is_empty_list_with_warning(
            "append_many_to_dataset", "json_content_array", json_content_array
        ):
            return None

        if status_array is not None:
            warnings.warn(
                (
                    "status_array is deprecated, asset status is automatically computed based on"
                    " its labels and cannot be overwritten."
                ),
                DeprecationWarning,
                stacklevel=1,
            )

        if content_array is None and json_content_array is None:
            raise ValueError("Variables content_array and json_content_array cannot be both None.")

        nb_data = (
            len(content_array)
            if content_array is not None
            else len(json_content_array)  # type:ignore
        )

        field_mapping = {
            "content": content_array,
            "json_content": json_content_array,
            "external_id": external_id_array,
            "id": id_array,
            "status": status_array,
            "json_metadata": json_metadata_array,
            "is_honeypot": is_honeypot_array,
        }
        assets = [{}] * nb_data
        for key, value in field_mapping.items():
            if value is not None:
                assets = [{**assets[i], key: value[i]} for i in range(nb_data)]
        result = import_assets(
            self.auth,
            project_id=project_id,
            assets=assets,
            disable_tqdm=disable_tqdm,
            verify=wait_until_availability,
        )
        return result

    @typechecked
    # pylint: disable=unused-argument
    def update_properties_in_assets(
        self,
        asset_ids: Optional[List[str]] = None,
        external_ids: Optional[List[str]] = None,
        priorities: Optional[List[int]] = None,
        json_metadatas: Optional[List[Union[dict, str]]] = None,
        consensus_marks: Optional[List[float]] = None,
        honeypot_marks: Optional[List[float]] = None,
        to_be_labeled_by_array: Optional[List[List[str]]] = None,
        contents: Optional[List[str]] = None,
        json_contents: Optional[List[str]] = None,
        status_array: Optional[List[str]] = None,
        is_used_for_consensus_array: Optional[List[bool]] = None,
        is_honeypot_array: Optional[List[bool]] = None,
        project_id: Optional[str] = None,
    ) -> List[Dict]:
        """Update the properties of one or more assets.

        Args:
            asset_ids: The internal asset IDs to modify.
            external_ids: The external asset IDs to modify (if `asset_ids` is not already provided).
            priorities: You can change the priority of the assets.
                By default, all assets have a priority of 0.
            json_metadatas: The metadata given to an asset should be stored
                in a json like dict with keys `imageUrl`, `text`, `url`:
                `json_metadata = {'imageUrl': '','text': '','url': ''}`
            consensus_marks: Should be between 0 and 1.
            honeypot_marks: Should be between 0 and 1.
            to_be_labeled_by_array: If given, each element of the list should contain the emails of
                the labelers authorized to label the asset.
            contents: - For a NLP project, the content can be directly in text format.
                - For an Image / Video / Pdf project, the content must be hosted on a web server,
                and you point Kili to your data by giving the URLs.
            json_contents: - For a NLP project, the `json_content`
                is a text formatted using RichText.
                - For a Video project, the`json_content` is a json containg urls pointing
                    to each frame of the video.
            status_array: Each element should be in `TODO`, `ONGOING`, `LABELED`,
                `TO_REVIEW`, `REVIEWED`.
            is_used_for_consensus_array: Whether to use the asset to compute consensus kpis or not.
            is_honeypot_array: Whether to use the asset for honeypot.
            project_id: The project ID. Only required if `external_ids` argument is provided.

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.

        Examples:
            >>> kili.update_properties_in_assets(
                    asset_ids=["ckg22d81r0jrg0885unmuswj8", "ckg22d81s0jrh0885pdxfd03n"],
                    consensus_marks=[1, 0.7],
                    contents=[None, 'https://to/second/asset.png'],
                    honeypot_marks=[0.8, 0.5],
                    is_honeypot_array=[True, True],
                    is_used_for_consensus_array=[True, False],
                    priorities=[None, 2],
                    status_array=['LABELED', 'REVIEWED'],
                    to_be_labeled_by_array=[['test+pierre@kili-technology.com'], None],
                )
        """
        if is_empty_list_with_warning(
            "update_properties_in_assets", "asset_ids", asset_ids
        ) or is_empty_list_with_warning(
            "update_properties_in_assets", "external_ids", external_ids
        ):
            return []

        if status_array is not None:
            warnings.warn(
                (
                    "status_array is deprecated, asset status is automatically computed based on"
                    " its labels and cannot be overwritten."
                ),
                DeprecationWarning,
                stacklevel=1,
            )

        if asset_ids is not None and external_ids is not None:
            warnings.warn(
                (
                    "The use of `external_ids` argument has changed. It is now used to identify"
                    " which properties of which assets to update. Please use"
                    " `kili.change_asset_external_ids()` method instead to change asset external"
                    " IDs."
                ),
                DeprecationWarning,
                stacklevel=1,
            )
            raise MissingArgumentError("Please provide either `asset_ids` or `external_ids`.")

        asset_ids = get_asset_ids_or_throw_error(self.auth, asset_ids, external_ids, project_id)

        saved_args = locals()
        parameters = {
            k: v
            for (k, v) in saved_args.items()
            if k
            in [
                "asset_ids",
                "priorities",
                "json_metadatas",
                "consensus_marks",
                "honeypot_marks",
                "to_be_labeled_by_array",
                "contents",
                "json_contents",
                "status_array",
                "is_used_for_consensus_array",
                "is_honeypot_array",
            ]
        }
        properties_to_batch = process_update_properties_in_assets_parameters(parameters)

        def generate_variables(batch: Dict) -> Dict:
            data = {
                "priority": batch["priorities"],
                "jsonMetadata": batch["json_metadatas"],
                "consensusMark": batch["consensus_marks"],
                "honeypotMark": batch["honeypot_marks"],
                "toBeLabeledBy": batch["to_be_labeled_by_array"],
                "shouldResetToBeLabeledBy": batch["should_reset_to_be_labeled_by_array"],
                "content": batch["contents"],
                "jsonContent": batch["json_contents"],
                "status": batch["status_array"],
                "isUsedForConsensus": batch["is_used_for_consensus_array"],
                "isHoneypot": batch["is_honeypot_array"],
            }
            data_array = [dict(zip(data, t)) for t in zip(*data.values())]
            return {
                "whereArray": [{"id": asset_id} for asset_id in batch["asset_ids"]],
                "dataArray": data_array,
            }

        results = _mutate_from_paginated_call(
            self,
            properties_to_batch,
            generate_variables,
            GQL_UPDATE_PROPERTIES_IN_ASSETS,
        )
        formated_results = [
            format_result("data", result, Asset, self.auth.ssl_verify) for result in results
        ]
        return [item for batch_list in formated_results for item in batch_list]

    @typechecked
    def change_asset_external_ids(
        self,
        new_external_ids: List[str],
        asset_ids: Optional[List[str]] = None,
        external_ids: Optional[List[str]] = None,
        project_id: Optional[str] = None,
    ) -> List[Dict]:
        """Update the external IDs of one or more assets.

        Args:
            new_external_ids: The new external IDs of the assets.
            asset_ids: The asset IDs to modify.
            external_ids: The external asset IDs to modify (if `asset_ids` is not already provided).
            project_id: The project ID. Only required if `external_ids` argument is provided.

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.

        Examples:
            >>> kili.change_asset_external_ids(
                    new_external_ids=["asset1", "asset2"],
                    asset_ids=["ckg22d81r0jrg0885unmuswj8", "ckg22d81s0jrh0885pdxfd03n"],
                )
        """
        if is_empty_list_with_warning(
            "change_asset_external_ids", "new_external_ids", new_external_ids
        ):
            return []

        asset_ids = get_asset_ids_or_throw_error(self.auth, asset_ids, external_ids, project_id)

        parameters = {
            "asset_ids": asset_ids,
            "new_external_ids": new_external_ids,
            "json_metadatas": None,
            "to_be_labeled_by_array": None,
        }
        properties_to_batch = process_update_properties_in_assets_parameters(parameters)

        def generate_variables(batch: Dict) -> Dict:
            data = {
                "externalId": batch["new_external_ids"],
                "jsonMetadata": batch["json_metadatas"],
                "toBeLabeledBy": batch["to_be_labeled_by_array"],
                "shouldResetToBeLabeledBy": batch["should_reset_to_be_labeled_by_array"],
            }
            data_array = [dict(zip(data, t)) for t in zip(*data.values())]
            return {
                "whereArray": [{"id": asset_id} for asset_id in batch["asset_ids"]],
                "dataArray": data_array,
            }

        results = _mutate_from_paginated_call(
            self,
            properties_to_batch,
            generate_variables,
            GQL_UPDATE_PROPERTIES_IN_ASSETS,
        )
        formated_results = [
            format_result("data", result, Asset, self.auth.ssl_verify) for result in results
        ]
        return [item for batch_list in formated_results for item in batch_list]

    @typechecked
    def delete_many_from_dataset(
        self,
        asset_ids: Optional[List[str]] = None,
        external_ids: Optional[List[str]] = None,
        project_id: Optional[str] = None,
    ) -> Asset:
        """Delete assets from a project.

        Args:
            asset_ids: The list of asset internal IDs to delete.
            external_ids: The list of asset external IDs to delete.
            project_id: The project ID. Only required if `external_ids` argument is provided.

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.
        """
        if is_empty_list_with_warning(
            "delete_many_from_dataset", "asset_ids", asset_ids
        ) or is_empty_list_with_warning("delete_many_from_dataset", "external_ids", external_ids):
            return Asset()

        asset_ids = get_asset_ids_or_throw_error(self.auth, asset_ids, external_ids, project_id)

        properties_to_batch: Dict[str, Optional[List[Any]]] = {"asset_ids": asset_ids}

        def generate_variables(batch):
            return {"where": {"idIn": batch["asset_ids"]}}

        @retry(
            wait=wait_exponential(multiplier=1, min=1, max=8),
            retry=retry_if_exception_type(MutationError),
            reraise=True,
        )
        def verify_last_batch(last_batch: Dict, results: List):
            """Check that all assets in the last batch have been deleted."""
            asset_ids = last_batch["asset_ids"][-1:]  # check last asset of the batch only
            nb_assets_in_kili = AssetQuery(self.auth.client, self.auth.ssl_verify).count(
                AssetWhere(
                    project_id=results[0]["data"]["id"],
                    asset_id_in=asset_ids,
                )
            )
            if nb_assets_in_kili > 0:
                raise MutationError("Failed to delete some assets.")

        results = _mutate_from_paginated_call(
            self,
            properties_to_batch,
            generate_variables,
            GQL_DELETE_MANY_FROM_DATASET,
            last_batch_callback=verify_last_batch,
        )
        return format_result("data", results[0], Asset, self.auth.ssl_verify)

    @typechecked
    def add_to_review(
        self,
        asset_ids: Optional[List[str]] = None,
        external_ids: Optional[List[str]] = None,
        project_id: Optional[str] = None,
    ) -> Optional[Dict[str, Any]]:
        """Add assets to review.

        !!! warning
            Assets without any label will be ignored.

        Args:
            asset_ids: The asset internal IDs to add to review.
            external_ids: The asset external IDs to add to review.
            project_id: The project ID. Only required if `external_ids` argument is provided.

        Returns:
            A dict object with the project `id` and the `asset_ids` of assets moved to review.
            `None` if no assets have changed status (already had `TO_REVIEW` status for example).
            An error message if mutation failed.

        Examples:
            >>> kili.add_to_review(
                    asset_ids=[
                        "ckg22d81r0jrg0885unmuswj8",
                        "ckg22d81s0jrh0885pdxfd03n",
                    ],
                )
        """
        if is_empty_list_with_warning(
            "add_to_review", "asset_ids", asset_ids
        ) or is_empty_list_with_warning("add_to_review", "external_ids", external_ids):
            return None

        asset_ids = get_asset_ids_or_throw_error(self.auth, asset_ids, external_ids, project_id)

        properties_to_batch: Dict[str, Optional[List[Any]]] = {"asset_ids": asset_ids}

        def generate_variables(batch):
            return {"where": {"idIn": batch["asset_ids"]}}

        @retry(
            wait=wait_exponential(multiplier=1, min=1, max=8),
            retry=retry_if_exception_type(MutationError),
            reraise=True,
        )
        def verify_last_batch(last_batch: Dict, results: List):
            """Check that all assets in the last batch have been sent to review."""
            try:
                project_id = results[0]["data"]["id"]
            except TypeError:
                return  # No assets have changed status
            asset_ids = last_batch["asset_ids"][-1:]  # check last asset of the batch only
            nb_assets_in_review = AssetQuery(self.auth.client, self.auth.ssl_verify).count(
                AssetWhere(
                    project_id=project_id,
                    asset_id_in=asset_ids,
                    status_in=["TO_REVIEW"],
                )
            )
            if len(asset_ids) != nb_assets_in_review:
                raise MutationError("Failed to send some assets to review")

        results = _mutate_from_paginated_call(
            self,
            properties_to_batch,
            generate_variables,
            GQL_ADD_ALL_LABELED_ASSETS_TO_REVIEW,
            last_batch_callback=verify_last_batch,
        )
        result = format_result("data", results[0], None, self.auth.ssl_verify)
        # unlike send_back_to_queue, the add_to_review mutation doesn't always return the project ID
        # it happens when no assets have been sent to review
        if isinstance(result, dict) and "id" in result:
            assets_in_review = AssetQuery(self.auth.client, self.auth.ssl_verify)(
                AssetWhere(project_id=result["id"], asset_id_in=asset_ids, status_in=["TO_REVIEW"]),
                ["id"],
                QueryOptions(disable_tqdm=True),
            )
            result["asset_ids"] = [asset["id"] for asset in assets_in_review]
            return result
        return result

    @typechecked
    def send_back_to_queue(
        self,
        asset_ids: Optional[List[str]] = None,
        external_ids: Optional[List[str]] = None,
        project_id: Optional[str] = None,
    ) -> Optional[Dict[str, Any]]:
        """Send assets back to queue.

        Args:
            asset_ids: List of internal IDs of assets to send back to queue.
            external_ids: List of external IDs of assets to send back to queue.
            project_id: The project ID. Only required if `external_ids` argument is provided.

        Returns:
            A dict object with the project `id` and the `asset_ids` of assets moved to queue.
            An error message if mutation failed.

        Examples:
            >>> kili.send_back_to_queue(
                    asset_ids=[
                        "ckg22d81r0jrg0885unmuswj8",
                        "ckg22d81s0jrh0885pdxfd03n",
                        ],
                )
        """
        if is_empty_list_with_warning(
            "send_back_to_queue", "asset_ids", asset_ids
        ) or is_empty_list_with_warning("send_back_to_queue", "external_ids", external_ids):
            return None

        asset_ids = get_asset_ids_or_throw_error(self.auth, asset_ids, external_ids, project_id)

        properties_to_batch: Dict[str, Optional[List[Any]]] = {"asset_ids": asset_ids}

        def generate_variables(batch):
            return {"where": {"idIn": batch["asset_ids"]}}

        @retry(
            wait=wait_exponential(multiplier=1, min=1, max=8),
            retry=retry_if_exception_type(MutationError),
            reraise=True,
        )
        def verify_last_batch(last_batch: Dict, results: List):
            """Check that all assets in the last batch have been sent back to queue."""
            asset_ids = last_batch["asset_ids"][-1:]  # check last asset of the batch only
            nb_assets_in_queue = AssetQuery(self.auth.client, self.auth.ssl_verify).count(
                AssetWhere(
                    project_id=results[0]["data"]["id"],
                    asset_id_in=asset_ids,
                    status_in=["ONGOING"],
                )
            )
            if len(asset_ids) != nb_assets_in_queue:
                raise MutationError("Failed to send some assets back to queue")

        results = _mutate_from_paginated_call(
            self,
            properties_to_batch,
            generate_variables,
            GQL_SEND_BACK_ASSETS_TO_QUEUE,
            last_batch_callback=verify_last_batch,
        )
        result = format_result("data", results[0], None, self.auth.ssl_verify)
        assets_in_queue = AssetQuery(self.auth.client, self.auth.ssl_verify)(
            AssetWhere(project_id=result["id"], asset_id_in=asset_ids, status_in=["ONGOING"]),
            ["id"],
            QueryOptions(disable_tqdm=True),
        )
        result["asset_ids"] = [asset["id"] for asset in assets_in_queue]
        return result
